# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from functools import cached_property
from typing import Any
from typing import List
from typing import Set
from typing import Tuple

import numpy as np
from anemoi.utils.dates import frequency_to_timedelta
from numpy.typing import NDArray

from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .debug import debug_indexing
from .forwards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import expand_list_indexing
from .indexing import index_to_slices
from .indexing import length_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class ConcatMixin:
    def __len__(self) -> int:
        """Returns the total length of the concatenated datasets.

        Returns
        -------
        int
            Total length of the concatenated datasets.
        """
        return sum(len(i) for i in self.datasets)

    @debug_indexing
    @expand_list_indexing
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Retrieves a tuple of data from the concatenated datasets based on the given index.

        Parameters
        ----------
        index : TupleIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Concatenated data array from the specified index.
        """
        index, changes = index_to_slices(index, self.shape)
        # print(index, changes)
        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(index[0], lengths)
        # print("slies", slices)
        result = [d[update_tuple(index, 0, i)[0]] for (d, i) in zip(self.datasets, slices) if i is not None]
        result = np.concatenate(result, axis=0)
        return apply_index_to_slices_changes(result, changes)

    @debug_indexing
    def __getitem__(self, n: FullIndex) -> NDArray[Any]:
        """Retrieves data from the concatenated datasets based on the given index.

        Parameters
        ----------
        n : FullIndex
            Index specifying the data to retrieve.

        Returns
        -------
        NDArray[Any]
            Data array from the concatenated datasets based on the index.
        """
        if isinstance(n, tuple):
            return self._get_tuple(n)

        if isinstance(n, slice):
            return self._get_slice(n)

        # TODO: optimize
        k = 0
        while n >= self.datasets[k]._len:
            n -= self.datasets[k]._len
            k += 1
        return self.datasets[k][n]

    @debug_indexing
    def _get_slice(self, s: slice) -> NDArray[Any]:
        """Retrieves a slice of data from the concatenated datasets.

        Parameters
        ----------
        s : slice
            Slice object specifying the range of data to retrieve.

        Returns
        -------
        NDArray[Any]
            Concatenated data array from the specified slice.
        """
        result = []

        lengths = [d.shape[0] for d in self.datasets]
        slices = length_to_slices(s, lengths)

        result = [d[i] for (d, i) in zip(self.datasets, slices) if i is not None]

        return np.concatenate(result)

    @cached_property
    def missing(self) -> Set[int]:
        """Returns the set of missing indices in the concatenated datasets."""
        result: Set[int] = set()
        offset = 0
        for d in self.datasets:
            result = result | set(m + offset for m in d.missing)
            offset += len(d)
        return result


class Concat(ConcatMixin, Combined):
    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Checks the compatibility of two datasets for concatenation.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        super().check_compatibility(d1, d2)
        self.check_same_sub_shapes(d1, d2, drop_axis=0)

    def check_same_lengths(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the lengths of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are concatenating along the first axis
        pass

    def check_same_dates(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the dates of two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        # Turned off because we are concatenating along the dates axis
        pass

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Returns the concatenated dates of all datasets."""
        return np.concatenate([d.dates for d in self.datasets])

    @property
    def shape(self) -> Shape:
        """Returns the shape of the concatenated datasets."""
        return (len(self),) + self.datasets[0].shape[1:]

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the concatenated datasets.

        Returns
        -------

        Node
            A Node object representing the concatenated datasets.
        """
        return Node(self, [d.tree() for d in self.datasets])

    @classmethod
    def check_dataset_compatibility(cls, datasets: List[Any], fill_missing_gaps: bool = False) -> List[Any]:
        """Checks the compatibility of the datasets for concatenation and fills missing gaps if required.

        Parameters
        ----------
        datasets : List[Any]
            List of datasets to check.
        fill_missing_gaps : bool, optional
            Whether to fill missing gaps between datasets, by default False.

        Returns
        -------
        List[Any]
            List of compatible datasets.
        """
        # Study the dates
        ranges = [(d.dates[0].astype(object), d.dates[-1].astype(object)) for d in datasets]

        # Make sure the dates are disjoint
        for i in range(len(ranges)):
            r = ranges[i]
            for j in range(i + 1, len(ranges)):
                s = ranges[j]
                if r[0] <= s[0] <= r[1] or r[0] <= s[1] <= r[1]:
                    raise ValueError(f"Overlapping dates: {r} and {s} ({datasets[i]} {datasets[j]})")

        # For now we should have the datasets in order with no gaps

        frequency = frequency_to_timedelta(datasets[0].frequency)
        result = []

        for i in range(len(ranges) - 1):
            result.append(datasets[i])
            r = ranges[i]
            s = ranges[i + 1]
            if r[1] + frequency != s[0]:
                if fill_missing_gaps:
                    from .missing import MissingDataset

                    result.append(MissingDataset(datasets[i], r[1] + frequency, s[0] - frequency))
                else:
                    r = [str(e) for e in r]
                    s = [str(e) for e in s]
                    raise ValueError(
                        "Datasets must be sorted by dates, with no gaps: "
                        f"{r} and {s} ({datasets[i]} {datasets[i+1]})"
                    )

        result.append(datasets[-1])
        assert len(result) >= len(datasets), (len(result), len(datasets))

        return result

    def forwards_subclass_metadata_specific(self) -> dict[str, Any]:
        """Get the metadata specific to the forwards subclass.

        Returns
        -------
        dict[str, Any]
            The metadata specific to the forwards subclass.
        """
        return {}


def concat_factory(args: Tuple[Any, ...], kwargs: dict) -> Concat:
    """Factory function to create a Concat object.

    Parameters
    ----------
    args : Tuple[Any, ...]
        Positional arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    Concat
        A Concat object.
    """
    datasets = kwargs.pop("concat")
    fill_missing_gaps = kwargs.pop("fill_missing_gaps", False)

    assert isinstance(datasets, (list, tuple))
    assert len(args) == 0

    assert isinstance(datasets, (list, tuple))

    datasets = [_open(e) for e in datasets]

    if len(datasets) == 1:
        return datasets[0]._subset(**kwargs)

    datasets, kwargs = _auto_adjust(datasets, kwargs)

    datasets = Concat.check_dataset_compatibility(datasets, fill_missing_gaps)

    return Concat(datasets)._subset(**kwargs)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from abc import ABC
from abc import abstractmethod
from typing import Any

import earthkit.data as ekd


class Filter(ABC):
    """A base class for filters."""

    def __init__(self, context: Any, *args: Any, **kwargs: Any) -> None:
        """Initialise the filter.

        Parameters
        ----------
        context : Any
            The context in which the filter is created.
        *args : tuple
            Positional arguments.
        **kwargs : dict
            Keyword arguments.
        """

        self.context = context

    @abstractmethod
    def execute(self, data: ekd.FieldList) -> ekd.FieldList:
        """Execute the filter.

        Parameters
        ----------
        data : ekd.FieldList
            The input data.

        Returns
        -------
        ekd.FieldList
            The output data.
        """

        pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
from functools import cache
from functools import wraps
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import Union
from unittest.mock import patch

import numpy as np
import pytest
import zarr
from anemoi.utils.dates import frequency_to_string
from anemoi.utils.dates import frequency_to_timedelta

from anemoi.datasets import open_dataset
from anemoi.datasets.data.concat import Concat
from anemoi.datasets.data.ensemble import Ensemble
from anemoi.datasets.data.grids import GridsBase
from anemoi.datasets.data.join import Join
from anemoi.datasets.data.misc import as_first_date
from anemoi.datasets.data.misc import as_last_date
from anemoi.datasets.data.select import Rename
from anemoi.datasets.data.select import Select
from anemoi.datasets.data.statistics import Statistics
from anemoi.datasets.data.stores import Zarr
from anemoi.datasets.data.subset import Subset

VALUES = 10


def mockup_open_zarr(func: Callable) -> Callable:
    """Decorator to mock the open_zarr function.

    Parameters
    ----------
    func : Callable
        Function to wrap.

    Returns
    -------
    Callable
        Wrapped function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        with patch("zarr.convenience.open", zarr_from_str):
            with patch("anemoi.datasets.data.stores.zarr_lookup", lambda name: name):
                return func(*args, **kwargs)

    return wrapper


@cache
def _(date: datetime.datetime, var: str, k: int = 0, e: int = 0, values: int = VALUES) -> np.ndarray:
    """Create a simple array of values based on the date, variable name, ensemble, grid, and other parameters.

    Parameters
    ----------
    date : datetime.datetime
        Date.
    var : str
        Variable name.
    k : int, optional
        Grid index, by default 0.
    e : int, optional
        Ensemble index, by default 0.
    values : int, optional
        Number of values, by default VALUES.

    Returns
    -------
    np.ndarray
        Array of values.
    """
    d = date.year * 10000 + date.month * 100 + date.day
    v = ord(var) - ord("a") + 1
    assert 0 <= k <= 9
    assert 0 <= e <= 9

    return np.array([d * 100 + v + k / 10.0 + w / 100.0 + e / 1000.0 for w in range(values)])


def create_zarr(
    vars: str = "abcd",
    start: int = 2021,
    end: int = 2021,
    frequency: datetime.timedelta = datetime.timedelta(hours=6),
    resolution: str = "o96",
    k: int = 0,
    ensemble: Optional[int] = None,
    grids: Optional[int] = None,
    missing: bool = False,
) -> zarr.Group:
    """Create a Zarr dataset.

    Parameters
    ----------
    vars : str, optional
        Variable names, by default "abcd".
    start : int, optional
        Start year, by default 2021.
    end : int, optional
        End year, by default 2021.
    frequency : datetime.timedelta, optional
        Frequency, by default datetime.timedelta(hours=6).
    resolution : str, optional
        Resolution, by default "o96".
    k : int, optional
        Grid index, by default 0.
    ensemble : Optional[int], optional
        Number of ensembles, by default None.
    grids : Optional[int], optional
        Number of grids, by default None.
    missing : bool, optional
        Whether to include missing dates, by default False.

    Returns
    -------
    zarr.Group
        Zarr dataset.
    """
    root = zarr.group()
    assert isinstance(frequency, datetime.timedelta)

    dates = []
    date = datetime.datetime(start, 1, 1)
    while date.year <= end:
        dates.append(date)
        date += frequency

    dates = np.array(dates, dtype="datetime64")

    ensembles = ensemble if ensemble is not None else 1
    values = grids if grids is not None else VALUES

    data = np.zeros(shape=(len(dates), len(vars), ensembles, values))

    for i, date in enumerate(dates):
        for j, var in enumerate(vars):
            for e in range(ensembles):
                data[i, j, e] = _(date.astype(object), var, k, e, values)

    root.create_dataset(
        "data",
        data=data,
        dtype=data.dtype,
        chunks=data.shape,
        compressor=None,
    )
    root.create_dataset(
        "dates",
        data=dates,
        compressor=None,
    )
    root.create_dataset(
        "latitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )
    root.create_dataset(
        "longitudes",
        data=np.array([x + values for x in range(values)]),
        compressor=None,
    )

    root.attrs["frequency"] = frequency_to_string(frequency)
    root.attrs["resolution"] = resolution
    root.attrs["name_to_index"] = {k: i for i, k in enumerate(vars)}

    root.attrs["data_request"] = {"grid": 1, "area": "g", "param_level": {}}
    root.attrs["variables_metadata"] = {v: {} for v in vars}

    if missing:
        missing_dates = []

        last = None
        for date in [d.astype(object) for d in dates]:
            yyyymmdd = date.strftime("%Y%m")
            if yyyymmdd != last:
                last = yyyymmdd
                missing_dates.append(date)

        root.attrs["missing_dates"] = [d.isoformat() for d in missing_dates]

    root.create_dataset(
        "mean",
        data=np.mean(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "stdev",
        data=np.std(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "maximum",
        data=np.max(data, axis=0),
        compressor=None,
    )
    root.create_dataset(
        "minimum",
        data=np.min(data, axis=0),
        compressor=None,
    )

    return root


def zarr_from_str(name: str, mode: str) -> zarr.Group:
    """Create a Zarr dataset from a string.

    Parameters
    ----------
    name : str
        Dataset name.
    mode : str
        Mode.

    Returns
    -------
    zarr.Group
        Zarr dataset.
    """
    # Format: test-2021-2021-6h-o96-abcd-0

    args = dict(
        test="test",
        start=2021,
        end=2021,
        frequency=6,
        resolution="o96",
        vars="abcd",
        k=0,
        ensemble=None,
        grids=None,
    )

    for name, bit in zip(args, name.split("-")):
        args[name] = bit

    print(args)

    return create_zarr(
        start=int(args["start"]),
        end=int(args["end"]),
        frequency=frequency_to_timedelta(args["frequency"]),
        resolution=args["resolution"],
        vars=[x for x in args["vars"]],
        k=int(args["k"]),
        ensemble=int(args["ensemble"]) if args["ensemble"] is not None else None,
        grids=int(args["grids"]) if args["grids"] is not None else None,
        missing=args["test"] == "missing",
    )


class IndexTester:
    """Class to test indexing of datasets."""

    def __init__(self, ds: Any) -> None:
        """Initialise the IndexTester.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        self.ds = ds
        self.np = ds[:]  # Numpy array

        assert self.ds.shape == self.np.shape
        assert (self.ds == self.np).all()

    def __getitem__(self, index: Any) -> None:
        """Test indexing.

        Parameters
        ----------
        index : Any
            Index.
        """
        print("INDEX", type(self.ds), index)
        if self.ds[index] is None:
            assert False, (self.ds, index)

        if not (self.ds[index] == self.np[index]).all():
            # print("DS", self.ds[index])
            # print("NP", self.np[index])
            assert (self.ds[index] == self.np[index]).all()


def make_row(*args: Any, ensemble: bool = False, grid: bool = False) -> np.ndarray:
    """Create a row of data.

    Parameters
    ----------
    *args : Any
        Additional arguments.
    ensemble : bool, optional
        Whether to include the ensemble dimension, by default False.
    grid : bool, optional
        Whether to include the grid dimension, by default False.

    Returns
    -------
    np.ndarray
        Row of data.
    """
    # assert not isinstance(args[0], (list, tuple))
    if grid:

        def _(x):
            return np.concatenate([np.array(p) for p in x])

        args = [_(a) for a in args]

    if ensemble:
        args = [np.array(a) for a in args]
    else:
        args = [np.array(a).reshape(1, -1) for a in args]  # Add ensemble dimension
    return np.array(args)


def make_missing(x: Any) -> Any:
    """Mark data as missing.

    Parameters
    ----------
    x : Any
        Data.

    Returns
    -------
    Any
        Data with missing values.
    """
    if isinstance(x, tuple):
        return (make_missing(a) for a in x)
    if isinstance(x, list):
        return [make_missing(a) for a in x]
    if isinstance(x, dict):
        return {k: make_missing(v) for k, v in x.items()}
    if isinstance(x, str) and x.startswith("test-"):
        return x.replace("test-", "missing-")
    return x


class DatasetTester:
    """Class to test various dataset operations."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise the DatasetTester.

        Parameters
        ----------
        *args : Any
            Arguments.
        **kwargs : Any
            Keyword arguments.
        """
        self.ds = open_dataset(*args, **kwargs)

        args, kwargs = make_missing((args, kwargs))

        print(f"ds={self.ds}")

    def run(
        self,
        *,
        expected_class: Type,
        expected_length: int,
        expected_shape: tuple,
        expected_variables: Union[str, list],
        expected_name_to_index: Union[str, dict],
        date_to_row: Callable,
        start_date: datetime.datetime,
        time_increment: datetime.timedelta,
        statistics_reference_dataset: Optional[Union[str, list]],
        statistics_reference_variables: Optional[Union[str, list]],
    ) -> None:
        """Run the dataset tests.

        Parameters
        ----------
        expected_class : Type
            Expected class.
        expected_length : int
            Expected length.
        expected_shape : tuple
            Expected shape.
        expected_variables : Union[str, list]
            Expected variables.
        expected_name_to_index : Union[str, dict]
            Expected name-to-index mapping.
        date_to_row : Callable
            Function to generate row data.
        start_date : datetime.datetime
            Start date.
        time_increment : datetime.timedelta
            Time increment.
        statistics_reference_dataset : Optional[Union[str, list]]
            Reference dataset for statistics.
        statistics_reference_variables : Optional[Union[str, list]]
            Reference variables for statistics.
        """
        if isinstance(expected_variables, str):
            expected_variables = [v for v in expected_variables]

        if isinstance(expected_name_to_index, str):
            expected_name_to_index = {v: i for i, v in enumerate(expected_name_to_index)}

        assert isinstance(self.ds, expected_class)
        assert len(self.ds) == expected_length
        assert len([row for row in self.ds]) == len(self.ds)
        assert self.ds.shape == expected_shape, (self.ds.shape, expected_shape)
        assert self.ds.variables == expected_variables

        assert set(self.ds.variables_metadata.keys()) == set(expected_variables)

        assert self.ds.name_to_index == expected_name_to_index
        assert self.ds.dates[0] == start_date
        assert self.ds.dates[1] - self.ds.dates[0] == time_increment

        dates = []
        date = start_date

        for row in self.ds:
            # print(f"{date=} {row.shape}")
            expect = date_to_row(date)
            assert (row == expect).all()
            dates.append(date)
            date += time_increment

        assert (self.ds.dates == np.array(dates, dtype="datetime64")).all()

        if statistics_reference_dataset is not None:
            self.same_stats(
                self.ds,
                open_dataset(statistics_reference_dataset),
                statistics_reference_variables,
            )

        self.indexing(self.ds)
        self.metadata(self.ds)

        self.ds.tree()

    def metadata(self, ds: Any) -> None:
        """Test metadata.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        metadata = ds.metadata()
        assert isinstance(metadata, dict)

    def same_stats(self, ds1: Any, ds2: Any, vars1: list, vars2: Optional[list] = None) -> None:
        """Compare statistics between two datasets.

        Parameters
        ----------
        ds1 : Any
            First dataset.
        ds2 : Any
            Second dataset.
        vars1 : list
            Variables in the first dataset.
        vars2 : Optional[list], optional
            Variables in the second dataset, by default None.
        """
        if vars2 is None:
            vars2 = vars1

        vars1 = [v for v in vars1]
        vars2 = [v for v in vars2]
        for v1, v2 in zip(vars1, vars2):
            idx1 = ds1.name_to_index[v1]
            idx2 = ds2.name_to_index[v2]
            assert (ds1.statistics["mean"][idx1] == ds2.statistics["mean"][idx2]).all()
            assert (ds1.statistics["stdev"][idx1] == ds2.statistics["stdev"][idx2]).all()
            assert (ds1.statistics["maximum"][idx1] == ds2.statistics["maximum"][idx2]).all()
            assert (ds1.statistics["minimum"][idx1] == ds2.statistics["minimum"][idx2]).all()

    def indexing(self, ds: Any) -> None:
        """Test indexing.

        Parameters
        ----------
        ds : Any
            Dataset.
        """
        t = IndexTester(ds)

        print("INDEXING", ds.shape)

        t[0:10, :, 0]
        t[:, 0:3, 0]
        # t[:, :, 0]
        t[0:10, 0:3, 0]
        t[:, :, :]

        if ds.shape[1] > 2:  # Variable dimension
            t[:, (1, 2), :]
            t[:, (1, 2)]

        t[0]
        t[0, :]
        t[0, 0, :]
        t[0, 0, 0, :]

        if ds.shape[2] > 1:  # Ensemble dimension
            t[0:10, :, (0, 1)]

        for i in range(3):
            t[i]
            start = 5 * i
            end = len(ds) - 5 * i
            step = len(ds) // 10

            t[start:end:step]
            t[start:end]
            t[start:]
            t[:end]
            t[::step]


def simple_row(date: datetime.datetime, vars: str) -> np.ndarray:
    """Create a simple row of data.

    Parameters
    ----------
    date : datetime.datetime
        Date.
    vars : str
        Variables.

    Returns
    -------
    np.ndarray
        Row of data.
    """
    values = [_(date, v) for v in vars]
    return make_row(*values)


@mockup_open_zarr
def test_simple() -> None:
    """Test a simple dataset."""
    test = DatasetTester("test-2021-2022-6h-o96-abcd")
    test.run(
        expected_class=Zarr,
        expected_length=365 * 2 * 4,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset="test-2021-2022-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_concat() -> None:
    """Test concatenating datasets."""
    test = DatasetTester(
        "test-2021-2022-6h-o96-abcd",
        "test-2023-2023-6h-o96-abcd",
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 3 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 3 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2022-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_join_1() -> None:
    """Test joining datasets (case 1)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", "test-2021-2021-6h-o96-efgh")
    test.run(
        expected_class=Join,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 8, 1, VALUES),
        expected_variables="abcdefgh",
        expected_name_to_index="abcdefgh",
        date_to_row=lambda date: simple_row(date, "abcdefgh"),
        # TODO: test second stats
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_join_2() -> None:
    """Test joining datasets (case 2)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd-1", "test-2021-2021-6h-o96-bdef-2")
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 6, 1, VALUES),
        expected_variables="abcdef",
        expected_name_to_index="abcdef",
        date_to_row=lambda date: make_row(
            _(date, "a", 1),
            _(date, "b", 2),
            _(date, "c", 1),
            _(date, "d", 2),
            _(date, "e", 2),
            _(date, "f", 2),
        ),
        statistics_reference_dataset=[
            "test-2021-2021-6h-o96-ac-1",
            "test-2021-2021-6h-o96-bdef-2",
        ],
        statistics_reference_variables="abcdef",
    )


@mockup_open_zarr
def test_join_3() -> None:
    """Test joining datasets (case 3)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd-1", "test-2021-2021-6h-o96-abcd-2")

    # TODO: This should trigger a warning about occulted dataset

    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            _(date, "a", 2),
            _(date, "b", 2),
            _(date, "c", 2),
            _(date, "d", 2),
        ),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd-2",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_1() -> None:
    """Test subsetting a dataset (case 1)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", frequency=12)
    test.run(
        expected_class=Subset,
        expected_length=365 * 3 * 2,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        expected_shape=(365 * 3 * 2, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_2() -> None:
    """Test subsetting a dataset (case 2)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=2022, end=2022)
    test.run(
        expected_class=Subset,
        expected_length=365 * 24,
        expected_shape=(365 * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2022, 1, 1),
        time_increment=datetime.timedelta(hours=1),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_3() -> None:
    """Test subsetting a dataset (case 3)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=2022, end=2022, frequency=12)
    test.run(
        expected_class=Subset,
        expected_length=365 * 2,
        expected_shape=(365 * 2, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2022, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_4() -> None:
    """Test subsetting a dataset (case 4)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=202206, end=202208)
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_5() -> None:
    """Test subsetting a dataset (case 5)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start=20220601, end=20220831)
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_6() -> None:
    """Test subsetting a dataset (case 6)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start="2022-06-01", end="2022-08-31")
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_7() -> None:
    """Test subsetting a dataset (case 7)."""
    test = DatasetTester("test-2021-2023-1h-o96-abcd", start="2022-06", end="2022-08")
    test.run(
        expected_class=Subset,
        expected_length=(30 + 31 + 31) * 24,
        start_date=datetime.datetime(2022, 6, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=((30 + 31 + 31) * 24, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2023-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_subset_8() -> None:
    """Test subsetting a dataset (case 8)."""
    test = DatasetTester(
        "test-2021-2021-1h-o96-abcd",
        start="03:00",
        frequency="6h",
    )
    test.run(
        expected_class=Subset,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1, 3, 0, 0),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-1h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_select_1() -> None:
    """Test selecting variables from a dataset (case 1)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select=["b", "d"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="bd",
        expected_name_to_index="bd",
        date_to_row=lambda date: simple_row(date, "bd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="bd",
    )


@mockup_open_zarr
def test_select_2() -> None:
    """Test selecting variables from a dataset (case 2)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select=["c", "a"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="ca",
        expected_name_to_index="ca",
        date_to_row=lambda date: simple_row(date, "ca"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="ca",
    )


@mockup_open_zarr
def test_select_3() -> None:
    """Test selecting variables from a dataset (case 3)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", select={"c", "a"})
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 2, 1, VALUES),
        expected_variables="ac",
        expected_name_to_index="ac",
        date_to_row=lambda date: simple_row(date, "ac"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="ac",
    )


@mockup_open_zarr
def test_rename() -> None:
    """Test renaming variables in a dataset."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", rename={"a": "x", "c": "y"})
    test.run(
        expected_class=Rename,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="xbyd",
        expected_name_to_index="xbyd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )

    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd"), "xbyd", "abcd")


@mockup_open_zarr
def test_drop() -> None:
    """Test dropping variables from a dataset."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", drop="a")
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 3, 1, VALUES),
        expected_variables="bcd",
        expected_name_to_index="bcd",
        date_to_row=lambda date: simple_row(date, "bcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="bcd",
    )


@mockup_open_zarr
def test_reorder_1() -> None:
    """Test reordering variables in a dataset (case 1)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", reorder=["d", "c", "b", "a"])
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="dcba",
        expected_name_to_index="dcba",
        date_to_row=lambda date: simple_row(date, "dcba"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="dcba",
    )


@mockup_open_zarr
def test_reorder_2() -> None:
    """Test reordering variables in a dataset (case 2)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd", reorder=dict(a=3, b=2, c=1, d=0))
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="dcba",
        expected_name_to_index="dcba",
        date_to_row=lambda date: simple_row(date, "dcba"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="dcba",
    )


@mockup_open_zarr
def test_constructor_1() -> None:
    """Test dataset constructor (case 1)."""
    ds1 = open_dataset("test-2021-2021-6h-o96-abcd")

    ds2 = open_dataset("test-2022-2022-6h-o96-abcd")

    test = DatasetTester(ds1, ds2)
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_2() -> None:
    """Test dataset constructor (case 2)."""
    test = DatasetTester(
        datasets=[
            "test-2021-2021-6h-o96-abcd",
            "test-2022-2022-6h-o96-abcd",
        ]
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_3() -> None:
    """Test dataset constructor (case 3)."""
    test = DatasetTester(
        {
            "datasets": [
                "test-2021-2021-6h-o96-abcd",
                "test-2022-2022-6h-o96-abcd",
            ]
        }
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_4() -> None:
    """Test dataset constructor (case 4)."""
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        {
            "dataset": "test-2022-2022-1h-o96-abcd",
            "frequency": 6,
        },
    )
    test.run(
        expected_class=Concat,
        expected_length=365 * 2 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 2 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_constructor_5() -> None:
    """Test dataset constructor (case 5)."""
    test = DatasetTester(
        {"dataset": "test-2021-2021-6h-o96-abcd-1", "rename": {"a": "x", "c": "y"}},
        {"dataset": "test-2021-2021-6h-o96-abcd-2", "rename": {"c": "z", "d": "t"}},
    )
    test.run(
        expected_class=Select,
        expected_length=365 * 4,
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 7, 1, VALUES),
        expected_variables="xbydazt",
        expected_name_to_index="xbydazt",
        date_to_row=lambda date: make_row(
            _(date, "a", 1),
            _(date, "b", 2),
            _(date, "c", 1),
            _(date, "d", 1),
            _(date, "a", 2),
            _(date, "c", 2),
            _(date, "d", 2),
        ),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )

    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd-1"), "xyd", "acd")
    test.same_stats(test.ds, open_dataset("test-2021-2021-6h-o96-abcd-2"), "abzt", "abcd")


@mockup_open_zarr
def test_dates() -> None:
    """Test date handling functions."""
    dates = None
    assert as_first_date(2021, dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date(2021, dates) == np.datetime64("2021-12-31T23:59:59")
    assert as_first_date("2021", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("2021", dates) == np.datetime64("2021-12-31T23:59:59")

    assert as_first_date(202106, dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date(202108, dates) == np.datetime64("2021-08-31T23:59:59")
    assert as_first_date("202106", dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date("202108", dates) == np.datetime64("2021-08-31T23:59:59")
    assert as_first_date("2021-06", dates) == np.datetime64("2021-06-01T00:00:00")
    assert as_last_date("2021-08", dates) == np.datetime64("2021-08-31T23:59:59")

    assert as_first_date(20210101, dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date(20210101, dates) == np.datetime64("2021-01-01T23:59:59")
    assert as_first_date("20210101", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("20210101", dates) == np.datetime64("2021-01-01T23:59:59")
    assert as_first_date("2021-01-01", dates) == np.datetime64("2021-01-01T00:00:00")
    assert as_last_date("2021-01-01", dates) == np.datetime64("2021-01-01T23:59:59")


@mockup_open_zarr
def test_dates_using_list() -> None:
    """Test date handling functions using a list of dates."""
    dates = [np.datetime64("2021-01-01T00:00:00") + i * np.timedelta64(6, "h") for i in range(3, 365 * 4 - 2)]
    assert dates[0] == np.datetime64("2021-01-01T18:00:00")
    assert dates[-1] == np.datetime64("2021-12-31T06:00:00")

    assert as_first_date(2021, dates) == np.datetime64("2021-01-01T18:00:00")
    assert as_last_date(2021, dates) == np.datetime64("2021-12-31T06:00:00")
    assert as_first_date("2021", dates) == np.datetime64("2021-01-01T18:00:00")
    assert as_last_date("2021", dates) == np.datetime64("2021-12-31T06:00:00")


@mockup_open_zarr
def test_dates_using_list_2() -> None:
    """Test date handling functions using a list of dates (case 2)."""
    dates = [np.datetime64("2021-01-01T00:00:00") + i * np.timedelta64(24, "h") for i in range(0, 10)]
    assert len(dates) == 10

    assert dates[0] == as_first_date("0%", dates)
    assert dates[0] == as_last_date("0%", dates)

    assert dates[0] == as_first_date("0.01%", dates)
    assert dates[0] == as_last_date("0.01%", dates)

    assert dates[0] == as_first_date("9.99%", dates)
    assert dates[0] == as_last_date("9.99%", dates)

    assert dates[0] == as_first_date("10%", dates)
    assert dates[0] == as_last_date("10%", dates)

    assert dates[1] == as_first_date("10.01%", dates)
    assert dates[0] == as_last_date("10.01%", dates)

    assert dates[1] == as_first_date("19.99%", dates)
    assert dates[0] == as_last_date("19.99%", dates)

    assert dates[1] == as_first_date("20%", dates)
    assert dates[1] == as_last_date("20%", dates)

    assert dates[2] == as_first_date("20.01%", dates)
    assert dates[1] == as_last_date("20.01%", dates)

    assert dates[-2] == as_first_date("89.99%", dates)
    assert dates[-3] == as_last_date("89.99%", dates)

    assert dates[-2] == as_first_date("90%", dates)
    assert dates[-2] == as_last_date("90%", dates)

    assert dates[-1] == as_first_date("90.01%", dates)
    assert dates[-2] == as_last_date("90.01%", dates)

    assert dates[-1] == as_first_date("99.99%", dates)
    assert dates[-2] == as_last_date("99.99%", dates)

    assert dates[-1] == as_first_date("100%", dates)
    assert dates[-1] == as_last_date("100%", dates)


@mockup_open_zarr
def test_slice_1() -> None:
    """Test slicing a dataset (case 1)."""
    test = DatasetTester("test-2021-2021-6h-o96-abcd")
    test.run(
        expected_class=Zarr,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_2() -> None:
    """Test slicing a dataset (case 2)."""
    test = DatasetTester([f"test-{year}-{year}-12h-o96-abcd" for year in range(1940, 2023)])
    test.run(
        expected_class=Concat,
        expected_length=60632,
        expected_shape=(60632, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(1940, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_3() -> None:
    """Test slicing a dataset (case 3)."""
    test = DatasetTester(
        [f"test-2020-2020-6h-o96-{vars}" for vars in ("abcd", "efgh", "ijkl", "mnop", "qrst", "uvwx", "yz")]
    )
    test.run(
        expected_class=Join,
        expected_length=366 * 4,
        date_to_row=lambda date: simple_row(date, "abcdefghijklmnopqrstuvwxyz"),
        start_date=datetime.datetime(2020, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(366 * 4, 26, 1, VALUES),
        expected_variables="abcdefghijklmnopqrstuvwxyz",
        expected_name_to_index="abcdefghijklmnopqrstuvwxyz",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_4() -> None:
    """Test slicing a dataset (case 4)."""
    test = DatasetTester([f"test-2020-2020-1h-o96-{vars}" for vars in ("abcd", "cd", "a", "c")])
    test.run(
        expected_class=Select,
        expected_length=8784,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2020, 1, 1),
        time_increment=datetime.timedelta(hours=1),
        expected_shape=(8784, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_slice_5() -> None:
    """Test slicing a dataset (case 5)."""
    test = DatasetTester(
        [f"test-{year}-{year}-6h-o96-abcd" for year in range(2010, 2020)],
        frequency=18,
    )
    test.run(
        expected_class=Subset,
        expected_length=4870,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2010, 1, 1),
        time_increment=datetime.timedelta(hours=18),
        expected_shape=(4870, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_1() -> None:
    """Test ensemble datasets (case 1)."""
    test = DatasetTester(
        ensemble=[
            "test-2021-2021-6h-o96-abcd-1-10",
            "test-2021-2021-6h-o96-abcd-2-1",
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 11, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_2() -> None:
    """Test ensemble datasets (case 2)."""
    test = DatasetTester(
        ensemble=[
            "test-2021-2021-6h-o96-abcd-1-10",
            "test-2021-2021-6h-o96-abcd-2-1",
            "test-2021-2021-6h-o96-abcd-3-5",
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 16, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)] + [_(date, "a", 3, i) for i in range(5)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)] + [_(date, "b", 3, i) for i in range(5)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)] + [_(date, "c", 3, i) for i in range(5)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)] + [_(date, "d", 3, i) for i in range(5)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_ensemble_3() -> None:
    """Test ensemble datasets (case 3)."""
    test = DatasetTester(
        ensemble=[
            {"dataset": "test-2021-2021-6h-o96-abcd-1-10", "frequency": 12},
            {"dataset": "test-2021-2021-6h-o96-abcd-2-1", "frequency": 12},
            {"dataset": "test-2021-2021-6h-o96-abcd-3-5", "frequency": 12},
        ]
    )
    test.run(
        expected_class=Ensemble,
        expected_length=365 * 1 * 2,
        expected_shape=(365 * 1 * 2, 4, 16, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [_(date, "a", 1, i) for i in range(10)] + [_(date, "a", 2, 0)] + [_(date, "a", 3, i) for i in range(5)],
            [_(date, "b", 1, i) for i in range(10)] + [_(date, "b", 2, 0)] + [_(date, "b", 3, i) for i in range(5)],
            [_(date, "c", 1, i) for i in range(10)] + [_(date, "c", 2, 0)] + [_(date, "c", 3, i) for i in range(5)],
            [_(date, "d", 1, i) for i in range(10)] + [_(date, "d", 2, 0)] + [_(date, "d", 3, i) for i in range(5)],
            ensemble=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=12),
        statistics_reference_dataset=None,
        statistics_reference_variables=None,
    )


@mockup_open_zarr
def test_grids() -> None:
    """Test datasets with different grids."""
    test = DatasetTester(
        grids=[
            "test-2021-2021-6h-o96-abcd-1-1",  # Default is 10 gridpoints
            "test-2021-2021-6h-o96-abcd-2-1-25",  # 25 gridpoints
        ]
    )
    test.run(
        expected_class=GridsBase,
        expected_length=365 * 1 * 4,
        expected_shape=(365 * 1 * 4, 4, 1, VALUES + 25),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        date_to_row=lambda date: make_row(
            [
                _(date, "a", 1),
                _(date, "a", 2, values=25),
            ],
            [
                _(date, "b", 1),
                _(date, "b", 2, values=25),
            ],
            [
                _(date, "c", 1),
                _(date, "c", 2, values=25),
            ],
            [
                _(date, "d", 1),
                _(date, "d", 2, values=25),
            ],
            grid=True,
        ),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        statistics_reference_dataset="test-2021-2021-6h-o96-abcd-1-1",
        statistics_reference_variables="abcd",
    )

    ds1 = open_dataset("test-2021-2021-6h-o96-abcd-1-1")
    ds2 = open_dataset("test-2021-2021-6h-o96-abcd-2-1-25")

    assert (test.ds.longitudes == np.concatenate([ds1.longitudes, ds2.longitudes])).all()
    assert (test.ds.latitudes == np.concatenate([ds1.latitudes, ds2.latitudes])).all()


@mockup_open_zarr
def test_statistics() -> None:
    """Test datasets with statistics."""
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        statistics="test-2000-2010-6h-o96-abcd",
    )
    test.run(
        expected_class=Statistics,
        expected_length=365 * 4,
        date_to_row=lambda date: simple_row(date, "abcd"),
        start_date=datetime.datetime(2021, 1, 1),
        time_increment=datetime.timedelta(hours=6),
        expected_shape=(365 * 4, 4, 1, VALUES),
        expected_variables="abcd",
        expected_name_to_index="abcd",
        statistics_reference_dataset="test-2000-2010-6h-o96-abcd",
        statistics_reference_variables="abcd",
    )


@mockup_open_zarr
def test_cropping() -> None:
    """Test cropping a dataset."""
    test = DatasetTester(
        "test-2021-2021-6h-o96-abcd",
        area=(18, 11, 11, 18),
    )
    assert test.ds.shape == (365 * 4, 4, 1, 8)


@mockup_open_zarr
def test_invalid_trim_edge() -> None:
    """Test that exception raised when attempting to trim a 1D dataset"""
    with pytest.raises(ValueError):
        DatasetTester(
            "test-2021-2021-6h-o96-abcd",
            trim_edge=(1, 2, 3, 4),
        )


if __name__ == "__main__":
    for name, obj in list(globals().items()):
        if name.startswith("test_") and callable(obj):
            print(f"Running {name}...")
            obj()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any

import earthkit.data as ekd
from anemoi.transform.fields import new_empty_fieldlist

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, **kwargs: Any) -> ekd.FieldList:
    """Create a pipeline that returns an empty result.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : List[Any]
        List of input fields.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        An empty result.
    """
    return new_empty_fieldlist()



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd

from anemoi.datasets.create.typing import DateList

from ..source import Source
from .xarray_support import XarrayFieldList
from .xarray_support import load_many
from .xarray_support import load_one

__all__ = ["load_many", "load_one", "XarrayFieldList"]


class XarraySourceBase(Source):
    """An Xarray base data source, intended to be subclassed."""

    emoji = "✖️"  # For tracing

    options: Optional[Dict[str, Any]] = None
    flavour: Optional[Dict[str, Any]] = None
    patch: Optional[Dict[str, Any]] = None

    path_or_url: Optional[str] = None

    def __init__(self, context: Any, path: str = None, url: str = None, *args: Any, **kwargs: Any):
        """Initialise the source.

        Parameters
        ----------
        context : Any
            The context for the data source.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, *args, **kwargs)

        if path is not None and url is not None:
            raise ValueError("Cannot specify both path and url")

        if path is not None:
            self.path_or_url = path
        else:
            self.path_or_url = url

        self.args = args
        self.kwargs = kwargs

    def execute(self, dates: DateList) -> ekd.FieldList:
        """Execute the data loading process for the given dates.

        Parameters
        ----------
        dates : DateList
            List of dates for which data needs to be loaded.

        Returns
        -------
        ekd.FieldList
            The loaded data fields.
        """

        # For now, just a simple wrapper around load_many
        # TODO: move the implementation here

        return load_many(
            self.emoji,
            self.context,
            dates,
            pattern=self.path_or_url,
            options=self.options,
            flavour=self.flavour,
            patch=self.patch,
            **self.kwargs,
        )


class XarraySource(XarraySourceBase):
    pass



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from collections import defaultdict
from typing import Any

import earthkit.data as ekd
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(context: Any, input: ekd.FieldList, wz: str, t: str, w: str = "w") -> ekd.FieldList:
    """Convert geometric vertical velocity (m/s) to vertical velocity (Pa / s).

    Parameters
    ----------
    context : Any
        The context for the execution.
    input : List[Any]
        The list of input fields.
    wz : str
        The parameter name for geometric vertical velocity.
    t : str
        The parameter name for temperature.
    w : str, optional
        The parameter name for vertical velocity. Defaults to "w".

    Returns
    -------
    ekd.FieldList
        The resulting FieldArray with converted vertical velocity fields.
    """
    result = []

    params = (wz, t)
    pairs = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")
        if param in params:
            key = tuple(key.items())

            if param in pairs[key]:
                raise ValueError(f"Duplicate field {param} for {key}")

            pairs[key][param] = f
            if param == t:
                result.append(f)
        else:
            result.append(f)

    for keys, values in pairs.items():

        if len(values) != 2:
            raise ValueError("Missing fields")

        wz_pl = values[wz].to_numpy(flatten=True)
        t_pl = values[t].to_numpy(flatten=True)
        pressure = next(
            float(v) * 100 for k, v in keys if k in ["level", "levelist"]
        )  # Looks first for "level" then "levelist" value
        w_pl = wz_to_w(wz_pl, t_pl, pressure)
        result.append(new_field_from_numpy(values[wz], w_pl, param=w))

    return new_fieldlist_from_list(result)


def wz_to_w(wz: Any, t: Any, pressure: float) -> Any:
    """Convert geometric vertical velocity (m/s) to vertical velocity (Pa / s).

    Parameters
    ----------
    wz : Any
        The geometric vertical velocity data.
    t : Any
        The temperature data.
    pressure : float
        The pressure value.

    Returns
    -------
    Any
        The vertical velocity data in Pa / s.
    """
    g = 9.81
    Rd = 287.058

    return -wz * g * pressure / (t * Rd)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import glob
import hashlib
import json
import logging
import os
import pickle
import shutil
import socket
from typing import Any
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import tqdm
from anemoi.utils.provenance import gather_provenance_info
from numpy.typing import NDArray

from ..check import check_data_values
from .summary import Summary

LOG = logging.getLogger(__name__)


def default_statistics_dates(dates: list[datetime.datetime]) -> tuple[datetime.datetime, datetime.datetime]:
    """Calculate default statistics dates based on the given list of dates.

    Parameters
    ----------
    dates : list of datetime.datetime
        List of datetime objects representing dates.

    Returns
    -------
    tuple of datetime.datetime
        A tuple containing the default start and end dates.
    """

    def to_datetime(d):
        if isinstance(d, np.datetime64):
            return d.tolist()
        assert isinstance(d, datetime.datetime), d
        return d

    first = dates[0]
    last = dates[-1]

    first = to_datetime(first)
    last = to_datetime(last)

    n_years = round((last - first).total_seconds() / (365.25 * 24 * 60 * 60))

    if n_years < 10:
        # leave out 20% of the data
        k = int(len(dates) * 0.8)
        end = dates[k - 1]
        LOG.info(f"Number of years {n_years} < 10, leaving out 20%. {end=}")
        return dates[0], end

    delta = 1
    if n_years >= 20:
        delta = 3
    LOG.info(f"Number of years {n_years}, leaving out {delta} years.")
    end_year = last.year - delta

    end = max(d for d in dates if to_datetime(d).year == end_year)
    return dates[0], end


def to_datetime(date: Union[str, datetime.datetime]) -> np.datetime64:
    """Convert a date to numpy datetime64 format.

    Parameters
    ----------
    date : str or datetime.datetime
        The date to convert.

    Returns
    -------
    numpy.datetime64
        The converted date.
    """
    if isinstance(date, str):
        return np.datetime64(date)
    if isinstance(date, datetime.datetime):
        return np.datetime64(date, "s")
    return date


def to_datetimes(dates: list[Union[str, datetime.datetime]]) -> list[np.datetime64]:
    """Convert a list of dates to numpy datetime64 format.

    Parameters
    ----------
    dates : list of str or datetime.datetime
        List of dates to convert.

    Returns
    -------
    list of numpy.datetime64
        List of converted dates.
    """
    return [to_datetime(d) for d in dates]


def fix_variance(x: float, name: str, count: NDArray[Any], sums: NDArray[Any], squares: NDArray[Any]) -> float:
    """Fix negative variance values due to numerical errors.

    Parameters
    ----------
    x : float
        The variance value.
    name : str
        The variable name.
    count : numpy.ndarray
        The count array.
    sums : numpy.ndarray
        The sums array.
    squares : numpy.ndarray
        The squares array.

    Returns
    -------
    float
        The fixed variance value.
    """
    assert count.shape == sums.shape == squares.shape
    assert isinstance(x, float)

    mean = sums / count
    assert mean.shape == count.shape

    if x >= 0:
        return x

    LOG.warning(f"Negative variance for {name=}, variance={x}")
    magnitude = np.sqrt((squares / count + mean * mean) / 2)
    LOG.warning(f"square / count - mean * mean =  {squares/count} - {mean*mean} = {squares/count - mean*mean}")
    LOG.warning(f"Variable span order of magnitude is {magnitude}.")
    LOG.warning(f"Count is {count}.")

    variances = squares / count - mean * mean
    assert variances.shape == squares.shape == mean.shape
    if np.all(variances >= 0):
        LOG.warning(f"All individual variances for {name} are positive, setting variance to 0.")
        return 0

    # if abs(x) < magnitude * 1e-6 and abs(x) < range * 1e-6:
    #     LOG.warning("Variance is negative but very small.")
    #     variances = squares / count - mean * mean
    #     return 0

    LOG.warning(f"ERROR at least one individual variance is negative ({np.nanmin(variances)}).")
    return 0


def check_variance(
    x: NDArray[Any],
    variables_names: list[str],
    minimum: NDArray[Any],
    maximum: NDArray[Any],
    mean: NDArray[Any],
    count: NDArray[Any],
    sums: NDArray[Any],
    squares: NDArray[Any],
) -> None:
    """Check for negative variance values and raise an error if found.

    Parameters
    ----------
    x : numpy.ndarray
        The variance array.
    variables_names : list of str
        List of variable names.
    minimum : numpy.ndarray
        The minimum values array.
    maximum : numpy.ndarray
        The maximum values array.
    mean : numpy.ndarray
        The mean values array.
    count : numpy.ndarray
        The count array.
    sums : numpy.ndarray
        The sums array.
    squares : numpy.ndarray
        The squares array.

    Raises
    ------
    ValueError
        If negative variance is found.
    """
    if (x >= 0).all():
        return
    print(x)
    print(variables_names)
    for i, (name, y) in enumerate(zip(variables_names, x)):
        if y >= 0:
            continue
        print("---")
        print(f"❗ Negative variance for {name=}, variance={y}")
        print(f" min={minimum[i]} max={maximum[i]} mean={mean[i]} count={count[i]} sums={sums[i]} squares={squares[i]}")
        print(f" -> sums: min={np.min(sums[i])}, max={np.max(sums[i])}, argmin={np.argmin(sums[i])}")
        print(f" -> squares: min={np.min(squares[i])}, max={np.max(squares[i])}, argmin={np.argmin(squares[i])}")
        print(f" -> count: min={np.min(count[i])}, max={np.max(count[i])}, argmin={np.argmin(count[i])}")
        print(
            f" squares / count - mean * mean =  {squares[i] / count[i]} - {mean[i] * mean[i]} = {squares[i] / count[i] - mean[i] * mean[i]}"
        )

    raise ValueError("Negative variance")


def compute_statistics(
    array: NDArray[Any], check_variables_names: Optional[List[str]] = None, allow_nans: bool = False
) -> dict[str, np.ndarray]:
    """Compute statistics for a given array, provides minimum, maximum, sum, squares, count and has_nans as a dictionary.

    Parameters
    ----------
    array : numpy.ndarray
        The array to compute statistics for.
    check_variables_names : list of str, optional
        List of variable names to check. Defaults to None.
    allow_nans : bool, optional
        Whether to allow NaN values. Defaults to False.

    Returns
    -------
    dict of str to numpy.ndarray
        A dictionary containing the computed statistics.
    """
    LOG.info(f"Computing statistics for {array.shape} array")
    nvars = array.shape[1]

    LOG.debug(f"Stats {nvars}, {array.shape}, {check_variables_names}")
    if check_variables_names:
        assert nvars == len(check_variables_names), (nvars, check_variables_names)
    stats_shape = (array.shape[0], nvars)

    count = np.zeros(stats_shape, dtype=np.int64)
    sums = np.zeros(stats_shape, dtype=np.float64)
    squares = np.zeros(stats_shape, dtype=np.float64)
    minimum = np.zeros(stats_shape, dtype=np.float64)
    maximum = np.zeros(stats_shape, dtype=np.float64)
    has_nans = np.zeros(stats_shape, dtype=np.bool_)

    for i, chunk in tqdm.tqdm(enumerate(array), delay=1, total=array.shape[0], desc="Computing statistics"):
        values = chunk.reshape((nvars, -1))

        for j, name in enumerate(check_variables_names):
            check_data_values(values[j, :], name=name, allow_nans=allow_nans)
            if np.isnan(values[j, :]).all():
                # LOG.warning(f"All NaN values for {name} ({j}) for date {i}")
                LOG.warning(f"All NaN values for {name} ({j}) for date {i}")

        # Ignore NaN values
        minimum[i] = np.nanmin(values, axis=1)
        maximum[i] = np.nanmax(values, axis=1)
        sums[i] = np.nansum(values, axis=1)
        squares[i] = np.nansum(values * values, axis=1)
        count[i] = np.sum(~np.isnan(values), axis=1)
        has_nans[i] = np.isnan(values).any()

    LOG.info(f"Statistics computed for {nvars} variables.")

    return {
        "minimum": minimum,
        "maximum": maximum,
        "sums": sums,
        "squares": squares,
        "count": count,
        "has_nans": has_nans,
    }


class TmpStatistics:
    """Temporary statistics storage class."""

    version = 3
    # Used in parrallel, during data loading,
    # to write statistics in pickled npz files.
    # can provide statistics for a subset of dates.

    def __init__(self, dirname: str, overwrite: bool = False) -> None:
        """Initialize TmpStatistics.

        Parameters
        ----------
        dirname : str
            Directory name for storing statistics.
        overwrite : bool, optional
            Whether to overwrite existing files. Defaults to False.
        """
        self.dirname = dirname
        self.overwrite = overwrite

    def add_provenance(self, **kwargs: dict) -> None:
        """Add provenance information.

        Parameters
        ----------
        **kwargs : dict
            Additional provenance information.
        """
        self.create(exist_ok=True)
        path = os.path.join(self.dirname, "provenance.json")
        if os.path.exists(path):
            return
        out = dict(provenance=gather_provenance_info(), **kwargs)
        with open(path, "w") as f:
            json.dump(out, f)

    def create(self, exist_ok: bool) -> None:
        """Create the directory for storing statistics.

        Parameters
        ----------
        exist_ok : bool
            Whether to ignore if the directory already exists.
        """
        os.makedirs(self.dirname, exist_ok=exist_ok)

    def delete(self) -> None:
        """Delete the directory for storing statistics."""
        try:
            shutil.rmtree(self.dirname)
        except FileNotFoundError:
            pass

    def write(self, key: str, data: any, dates: list[datetime.datetime]) -> None:
        """Write statistics data to a file.

        Parameters
        ----------
        key : str
            The key for the data.
        data : any
            The data to write.
        dates : list of datetime.datetime
            List of dates associated with the data.
        """
        self.create(exist_ok=True)
        h = hashlib.sha256(str(dates).encode("utf-8")).hexdigest()
        path = os.path.join(self.dirname, f"{h}.npz")

        if not self.overwrite:
            assert not os.path.exists(path), f"{path} already exists"

        tmp_path = path + f".tmp-{os.getpid()}-on-{socket.gethostname()}"
        with open(tmp_path, "wb") as f:
            pickle.dump((key, dates, data), f)
        shutil.move(tmp_path, path)

        LOG.debug(f"Written statistics data for {len(dates)} dates in {path} ({dates})")

    def _gather_data(self) -> tuple[str, list[datetime.datetime], dict]:
        """Gather data from stored files.

        Yields
        ------
        tuple of str, list of datetime.datetime, dict
            A tuple containing key, dates, and data.
        """
        # use glob to read all pickles
        files = glob.glob(self.dirname + "/*.npz")
        LOG.debug(f"Reading stats data, found {len(files)} files in {self.dirname}")
        assert len(files) > 0, f"No files found in {self.dirname}"
        for f in files:
            with open(f, "rb") as f:
                yield pickle.load(f)

    def get_aggregated(self, *args: Any, **kwargs: Any) -> Summary:
        """Get aggregated statistics.

        Parameters
        ----------
        *args : Any
            Additional arguments.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Summary
            The aggregated statistics summary.
        """
        aggregator = StatAggregator(self, *args, **kwargs)
        return aggregator.aggregate()

    def __str__(self) -> str:
        """String representation of TmpStatistics.

        Returns
        -------
        str
            The string representation.
        """
        return f"TmpStatistics({self.dirname})"


class StatAggregator:
    """Statistics aggregator class."""

    NAMES = ["minimum", "maximum", "sums", "squares", "count", "has_nans"]

    def __init__(
        self, owner: TmpStatistics, dates: list[datetime.datetime], variables_names: list[str], allow_nans: bool
    ) -> None:
        """Initialize StatAggregator.

        Parameters
        ----------
        owner : TmpStatistics
            The owner TmpStatistics instance.
        dates : list of datetime.datetime
            List of dates to aggregate.
        variables_names : list of str
            List of variable names.
        allow_nans : bool
            Whether to allow NaN values.
        """
        dates = sorted(dates)
        dates = to_datetimes(dates)
        assert dates, "No dates selected"
        self.owner = owner
        self.dates = dates
        self._number_of_dates = len(dates)
        self._set_of_dates = set(dates)
        self.variables_names = variables_names
        self.allow_nans = allow_nans

        self.shape = (self._number_of_dates, len(self.variables_names))
        LOG.debug(f"Aggregating statistics on shape={self.shape}. Variables : {self.variables_names}")

        self.minimum = np.full(self.shape, np.nan, dtype=np.float64)
        self.maximum = np.full(self.shape, np.nan, dtype=np.float64)
        self.sums = np.full(self.shape, np.nan, dtype=np.float64)
        self.squares = np.full(self.shape, np.nan, dtype=np.float64)
        self.count = np.full(self.shape, -1, dtype=np.int64)
        self.has_nans = np.full(self.shape, False, dtype=np.bool_)

        self._read()

    def _read(self) -> None:
        """Read and aggregate statistics data from files."""

        def check_type(a, b):
            if not isinstance(a, set):
                a = set(list(a))
            if not isinstance(b, set):
                b = set(list(b))
            a = next(iter(a)) if a else None
            b = next(iter(b)) if b else None
            assert type(a) is type(b), (type(a), type(b))

        found = set()
        offset = 0

        for _, _dates, stats in self.owner._gather_data():
            assert isinstance(stats, dict), stats
            assert stats["minimum"].shape[0] == len(_dates), (stats["minimum"].shape, len(_dates))
            assert stats["minimum"].shape[1] == len(self.variables_names), (
                stats["minimum"].shape,
                len(self.variables_names),
            )
            for n in self.NAMES:
                assert n in stats, (n, list(stats.keys()))
            _dates = to_datetimes(_dates)
            check_type(_dates, self._set_of_dates)
            if found:
                check_type(found, self._set_of_dates)
                assert found.isdisjoint(_dates), "Duplicate dates found in precomputed statistics"

            # filter dates
            dates = set(_dates) & self._set_of_dates

            if not dates:
                # dates have been completely filtered for this chunk
                continue

            # filter data
            bitmap = np.array([d in self._set_of_dates for d in _dates])
            for k in self.NAMES:
                stats[k] = stats[k][bitmap]

            assert stats["minimum"].shape[0] == len(dates), (stats["minimum"].shape, len(dates))

            # store data in self
            found |= set(dates)
            for name in self.NAMES:
                array = getattr(self, name)
                assert stats[name].shape[0] == len(dates), (stats[name].shape, len(dates))
                array[offset : offset + len(dates)] = stats[name]
            offset += len(dates)

        for d in self.dates:
            assert d in found, f"Statistics for date {d} not precomputed."
        assert self._number_of_dates == len(found), "Not all dates found in precomputed statistics"
        assert self._number_of_dates == offset, "Not all dates found in precomputed statistics."
        LOG.debug(f"Statistics for {len(found)} dates found.")

    def aggregate(self) -> Summary:
        """Aggregate the statistics data.

        Returns
        -------
        Summary
            The aggregated statistics summary.
        """
        minimum = np.nanmin(self.minimum, axis=0)
        maximum = np.nanmax(self.maximum, axis=0)

        sums = np.nansum(self.sums, axis=0)
        squares = np.nansum(self.squares, axis=0)
        count = np.nansum(self.count, axis=0)
        has_nans = np.any(self.has_nans, axis=0)
        assert sums.shape == count.shape == squares.shape == minimum.shape == maximum.shape

        mean = sums / count
        assert mean.shape == minimum.shape

        x = squares / count - mean * mean
        assert x.shape == minimum.shape

        for i, name in enumerate(self.variables_names):
            # remove negative variance due to numerical errors
            x[i] = fix_variance(x[i], name, self.count[i : i + 1], self.sums[i : i + 1], self.squares[i : i + 1])

        for i, name in enumerate(self.variables_names):
            check_variance(
                x[i : i + 1],
                [name],
                minimum[i : i + 1],
                maximum[i : i + 1],
                mean[i : i + 1],
                count[i : i + 1],
                sums[i : i + 1],
                squares[i : i + 1],
            )
            check_data_values(np.array([mean[i]]), name=name, allow_nans=False)

        stdev = np.sqrt(x)

        return Summary(
            minimum=minimum,
            maximum=maximum,
            mean=mean,
            count=count,
            sums=sums,
            squares=squares,
            stdev=stdev,
            variables_names=self.variables_names,
            has_nans=has_nans,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
from functools import cached_property
from functools import wraps
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

import numpy as np
from numpy.typing import NDArray

from .concat import ConcatMixin
from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .debug import Node
from .forwards import Combined
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class check:
    """A decorator class to perform checks before calling a method."""

    def __init__(self, check: str) -> None:
        """Initialize the check decorator.

        Parameters
        ----------
        check : str
            The name of the check method.
        """
        self.check = check

    def __call__(self, method: Callable) -> Callable:
        """Call the check decorator.

        Parameters
        ----------
        method : Callable
            The method to decorate.

        Returns
        -------
        Callable
            The decorated method.
        """
        name = method.__name__
        check = self.check

        @wraps(method)
        def wrapper(obj: "Unchecked") -> Any:
            """Wrapper function to check compatibility before calling the method.

            Parameters
            ----------
            obj : Unchecked
                The Unchecked object.

            Returns
            -------
            Any
                The result of the method.
            """
            for d in obj.datasets[1:]:
                getattr(obj, check)(obj.datasets[0], d)

            return getattr(Combined, name).__get__(obj)

        return wrapper


class Unchecked(Combined):
    """A class representing a dataset without compatibility checks."""

    def tree(self) -> Node:
        """Get the tree representation of the dataset.

        Returns
        -------
        Node
            The tree representation.
        """
        return Node(self, [d.tree() for d in self.datasets])

    def _subset(self, **kwargs: dict) -> "Unchecked":
        """Get a subset of the dataset.

        Parameters
        ----------
        **kwargs : dict
            Subset parameters.

        Returns
        -------
        Unchecked
            The subset of the dataset.
        """
        assert not kwargs
        return self

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Check compatibility between two datasets.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        pass

    @property
    @check("check_same_dates")
    def dates(self) -> NDArray[np.datetime64]:
        """Get the dates of the dataset."""
        pass

    @property
    @check("check_same_resolution")
    def resolution(self) -> Any:
        """Get the resolution of the dataset."""
        pass

    @property
    def field_shape(self) -> tuple:
        """Get the field shape of the dataset."""
        raise NotImplementedError()

    @property
    @check("check_same_frequency")
    def frequency(self) -> datetime.timedelta:
        """Get the frequency of the dataset."""
        raise NotImplementedError()

    @property
    @check("check_same_grid")
    def latitudes(self) -> NDArray[Any]:
        """Get the latitudes of the dataset."""
        raise NotImplementedError()

    @property
    @check("check_same_grid")
    def longitudes(self) -> NDArray[Any]:
        """Get the longitudes of the dataset."""
        raise NotImplementedError()

    @check("check_same_variables")
    @property
    def name_to_index(self) -> Dict[str, int]:
        """Get the mapping of variable names to their indices."""
        raise NotImplementedError()

    @check("check_same_variables")
    @property
    def variables(self) -> List[str]:
        """Get the list of variables in the dataset."""
        raise NotImplementedError()

    @check("check_same_variables")
    @property
    def variables_metadata(self) -> dict:
        """Get the metadata for the variables."""
        raise NotImplementedError()

    @check("check_same_variables")
    @property
    def statistics(self) -> Dict[str, NDArray[Any]]:
        """Get the statistics of the dataset."""
        raise NotImplementedError()

    @check("check_same_variables")
    def statistics_tendencies(self, delta: Optional[datetime.timedelta] = None) -> Dict[str, NDArray[Any]]:
        """Get the statistics tendencies of the dataset.

        Parameters
        ----------
        delta : Optional[datetime.timedelta]
            The time delta for tendencies.

        Returns
        -------
        Dict[str, NDArray[Any]]
            The statistics tendencies.
        """
        raise NotImplementedError()

    @property
    def shape(self) -> Shape:
        """Get the shape of the dataset."""
        raise NotImplementedError()

    @cached_property
    def missing(self) -> Set[int]:
        """Get the missing data indices."""
        result: Set[int] = set()
        for d in self.datasets:
            result = result | d.missing
        return result


class Chain(ConcatMixin, Unchecked):
    """A class representing a chain of datasets without compatibility checks."""

    def __len__(self) -> int:
        """Get the length of the dataset.

        Returns
        -------
        int
            The length of the dataset.
        """
        return sum(len(d) for d in self.datasets)

    def __getitem__(self, n: FullIndex) -> tuple:
        """Get an item from the dataset.

        Parameters
        ----------
        n : FullIndex
            The index of the item.

        Returns
        -------
        tuple
            The item at the specified index.
        """
        return tuple(d[n] for d in self.datasets)

    @property
    def dates(self) -> NDArray[np.datetime64]:
        """Get the dates of the dataset."""
        raise NotImplementedError()

    def dataset_metadata(self) -> dict:
        """Get the metadata of the dataset.

        Returns
        -------
        dict
            The metadata of the dataset.
        """
        return {"multiple": [d.dataset_metadata() for d in self.datasets]}


def chain_factory(args: tuple, kwargs: dict) -> Dataset:
    """Factory function to create a Chain dataset.

    Parameters
    ----------
    args : tuple
        Positional arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    Dataset
        The Chain dataset.
    """
    chain = kwargs.pop("chain")
    assert len(args) == 0
    assert isinstance(chain, (list, tuple))

    datasets = [_open(e) for e in chain]
    datasets, kwargs = _auto_adjust(datasets, kwargs)

    return Chain(datasets)._subset(**kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import base64
import logging
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

LOG = logging.getLogger(__name__)


def plot_mask(
    path: str,
    mask: NDArray[Any],
    lats: NDArray[Any],
    lons: NDArray[Any],
    global_lats: NDArray[Any],
    global_lons: NDArray[Any],
) -> None:
    """Plot and save various visualizations of the mask and coordinates.

    Parameters
    ----------
    path : str
        The base path for saving the plots.
    mask : NDArray[Any]
        The mask array.
    lats : NDArray[Any]
        Latitude coordinates.
    lons : NDArray[Any]
        Longitude coordinates.
    global_lats : NDArray[Any]
        Global latitude coordinates.
    global_lons : NDArray[Any]
        Global longitude coordinates.
    """
    import matplotlib.pyplot as plt

    s = 1

    global_lons[global_lons >= 180] -= 360

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons, global_lats, s=s, marker="o", c="r")
    if isinstance(path, str):
        plt.savefig(path + "-global.png")

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="k")
    if isinstance(path, str):
        plt.savefig(path + "-cutout.png")

    plt.figure(figsize=(10, 5))
    plt.scatter(lons, lats, s=s)
    if isinstance(path, str):
        plt.savefig(path + "-lam.png")
    # plt.scatter(lons, lats, s=0.01)

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="r")
    plt.scatter(lons, lats, s=s)
    if isinstance(path, str):
        plt.savefig(path + "-both.png")
    # plt.scatter(lons, lats, s=0.01)

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="r")
    plt.scatter(lons, lats, s=s)
    plt.xlim(np.amin(lons) - 1, np.amax(lons) + 1)
    plt.ylim(np.amin(lats) - 1, np.amax(lats) + 1)
    if isinstance(path, str):
        plt.savefig(path + "-both-zoomed.png")

    plt.figure(figsize=(10, 5))
    plt.scatter(global_lons[mask], global_lats[mask], s=s, c="r")
    plt.xlim(np.amin(lons) - 1, np.amax(lons) + 1)
    plt.ylim(np.amin(lats) - 1, np.amax(lats) + 1)
    if isinstance(path, str):
        plt.savefig(path + "-global-zoomed.png")


# TODO: Use the one from anemoi.utils.grids instead
# from anemoi.utils.grids import ...
def xyz_to_latlon(x: NDArray[Any], y: NDArray[Any], z: NDArray[Any]) -> Tuple[NDArray[Any], NDArray[Any]]:
    """Convert Cartesian coordinates to latitude and longitude.

    Parameters
    ----------
    x : NDArray[Any]
        X coordinates.
    y : NDArray[Any]
        Y coordinates.
    z : NDArray[Any]
        Z coordinates.

    Returns
    -------
    Tuple[NDArray[Any], NDArray[Any]]
        Latitude and longitude coordinates.
    """
    return (
        np.rad2deg(np.arcsin(np.minimum(1.0, np.maximum(-1.0, z)))),
        np.rad2deg(np.arctan2(y, x)),
    )


# TODO: Use the one from anemoi.utils.grids instead
# from anemoi.utils.grids import ...
def latlon_to_xyz(
    lat: NDArray[Any], lon: NDArray[Any], radius: float = 1.0
) -> Tuple[NDArray[Any], NDArray[Any], NDArray[Any]]:
    """Convert latitude and longitude to Cartesian coordinates.

    Parameters
    ----------
    lat : NDArray[Any]
        Latitude coordinates.
    lon : NDArray[Any]
        Longitude coordinates.
    radius : float, optional
        Radius of the sphere. Defaults to 1.0.

    Returns
    -------
    Tuple[NDArray[Any], NDArray[Any], NDArray[Any]]
        X, Y, and Z coordinates.
    """
    # https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
    # We assume that the Earth is a sphere of radius 1 so N(phi) = 1
    # We assume h = 0
    #
    phi = np.deg2rad(lat)
    lda = np.deg2rad(lon)

    cos_phi = np.cos(phi)
    cos_lda = np.cos(lda)
    sin_phi = np.sin(phi)
    sin_lda = np.sin(lda)

    x = cos_phi * cos_lda * radius
    y = cos_phi * sin_lda * radius
    z = sin_phi * radius

    return x, y, z


class Triangle3D:
    """A class to represent a 3D triangle and perform intersection tests with rays."""

    def __init__(self, v0: NDArray[Any], v1: NDArray[Any], v2: NDArray[Any]) -> None:
        """Initialize the Triangle3D object.

        Parameters
        ----------
        v0 : NDArray[Any]
            First vertex of the triangle.
        v1 : NDArray[Any]
            Second vertex of the triangle.
        v2 : NDArray[Any]
            Third vertex of the triangle.
        """
        self.v0 = v0
        self.v1 = v1
        self.v2 = v2

    def intersect(self, ray_origin: NDArray[Any], ray_direction: NDArray[Any]) -> bool:
        """Check if a ray intersects with the triangle.

        Parameters
        ----------
        ray_origin : NDArray[Any]
            Origin of the ray.
        ray_direction : NDArray[Any]
            Direction of the ray.

        Returns
        -------
        bool
            True if the ray intersects with the triangle, False otherwise.
        """
        # Möller–Trumbore intersection algorithm
        # https://en.wikipedia.org/wiki/M%C3%B6ller%E2%80%93Trumbore_intersection_algorithm

        epsilon = 0.0000001

        h = np.cross(ray_direction, self.v2 - self.v0)
        a = np.dot(self.v1 - self.v0, h)

        if -epsilon < a < epsilon:
            return False

        f = 1.0 / a
        s = ray_origin - self.v0
        u = f * np.dot(s, h)

        if u < 0.0 or u > 1.0:
            return False

        q = np.cross(s, self.v1 - self.v0)
        v = f * np.dot(ray_direction, q)

        if v < 0.0 or u + v > 1.0:
            return False

        t = f * np.dot(self.v2 - self.v0, q)

        if t > epsilon:
            return True

        return False


def cropping_mask(
    lats: NDArray[Any],
    lons: NDArray[Any],
    north: float,
    west: float,
    south: float,
    east: float,
) -> NDArray[Any]:
    """Create a mask for the points within the specified latitude and longitude bounds.

    Parameters
    ----------
    lats : NDArray[Any]
        Latitude coordinates.
    lons : NDArray[Any]
        Longitude coordinates.
    north : float
        Northern boundary.
    west : float
        Western boundary.
    south : float
        Southern boundary.
    east : float
        Eastern boundary.

    Returns
    -------
    NDArray[Any]
        Mask array.
    """
    mask = (
        (lats >= south)
        & (lats <= north)
        & (
            ((lons >= west) & (lons <= east))
            | ((lons >= west + 360) & (lons <= east + 360))
            | ((lons >= west - 360) & (lons <= east - 360))
        )
    )
    return mask


def cutout_mask(
    lats: NDArray[Any],
    lons: NDArray[Any],
    global_lats: NDArray[Any],
    global_lons: NDArray[Any],
    cropping_distance: float = 2.0,
    neighbours: int = 5,
    min_distance_km: Optional[Union[int, float]] = None,
    plot: Optional[str] = None,
) -> NDArray[Any]:
    """Return a mask for the points in [global_lats, global_lons] that are inside of [lats, lons].

    Parameters
    ----------
    lats : NDArray[Any]
        Latitude coordinates.
    lons : NDArray[Any]
        Longitude coordinates.
    global_lats : NDArray[Any]
        Global latitude coordinates.
    global_lons : NDArray[Any]
        Global longitude coordinates.
    cropping_distance : float, optional
        Cropping distance. Defaults to 2.0.
    neighbours : int, optional
        Number of neighbours. Defaults to 5.
    min_distance_km : Optional[Union[int, float]], optional
        Minimum distance in kilometers. Defaults to None.
    plot : Optional[str], optional
        Path for saving the plot. Defaults to None.

    Returns
    -------
    NDArray[Any]
        Mask array.
    """
    from scipy.spatial import cKDTree

    # TODO: transform min_distance from lat/lon to xyz

    assert global_lats.ndim == 1
    assert global_lons.ndim == 1
    assert lats.ndim == 1
    assert lons.ndim == 1

    assert global_lats.shape == global_lons.shape
    assert lats.shape == lons.shape

    north = np.amax(lats)
    south = np.amin(lats)
    east = np.amax(lons)
    west = np.amin(lons)

    # Reduce the global grid to the area of interest

    mask = cropping_mask(
        global_lats,
        global_lons,
        np.min([90.0, north + cropping_distance]),
        west - cropping_distance,
        np.max([-90.0, south - cropping_distance]),
        east + cropping_distance,
    )

    # return mask
    # mask = np.array([True] * len(global_lats), dtype=bool)
    global_lats_masked = global_lats[mask]
    global_lons_masked = global_lons[mask]

    global_xyx = latlon_to_xyz(global_lats_masked, global_lons_masked)
    global_points = np.array(global_xyx).transpose()

    xyx = latlon_to_xyz(lats, lons)
    lam_points = np.array(xyx).transpose()

    if isinstance(min_distance_km, (int, float)):
        min_distance = min_distance_km / 6371.0
    else:
        points = {"lam": lam_points, "global": global_points, None: global_points}[min_distance_km]
        distances, _ = cKDTree(points).query(points, k=2)
        min_distance = np.min(distances[:, 1])

        LOG.info(f"cutout_mask using min_distance = {min_distance * 6371.0} km")

    # Use a cKDTree to find the nearest points
    distances, indices = cKDTree(lam_points).query(global_points, k=neighbours)

    # Centre of the Earth
    zero = np.array([0.0, 0.0, 0.0])

    # After the loop, 'inside_lam' will contain a list point to EXCLUDE
    inside_lam = []

    for i, (global_point, distance, index) in enumerate(zip(global_points, distances, indices)):

        # We check more than one triangle in case te global point
        # is near the edge of triangle, (the lam point and global points are colinear)

        inside = False
        for j in range(neighbours):
            t = Triangle3D(
                lam_points[index[j]], lam_points[index[(j + 1) % neighbours]], lam_points[index[(j + 2) % neighbours]]
            )
            inside = t.intersect(zero, global_point)
            if inside:
                break

        close = np.min(distance) <= min_distance

        inside_lam.append(inside or close)

    j = 0
    inside_lam_array = np.array(inside_lam)
    for i, m in enumerate(mask):
        if not m:
            continue

        mask[i] = inside_lam_array[j]
        j += 1

    assert j == len(inside_lam_array)

    # Invert the mask, so we have only the points outside the cutout
    mask = ~mask

    if plot:
        plot_mask(plot, mask, lats, lons, global_lats, global_lons)

    return mask


def thinning_mask(
    lats: NDArray[Any],
    lons: NDArray[Any],
    global_lats: NDArray[Any],
    global_lons: NDArray[Any],
    cropping_distance: float = 2.0,
) -> NDArray[Any]:
    """Return the list of points in [lats, lons] closest to [global_lats, global_lons].

    Parameters
    ----------
    lats : NDArray[Any]
        Latitude coordinates.
    lons : NDArray[Any]
        Longitude coordinates.
    global_lats : NDArray[Any]
        Global latitude coordinates.
    global_lons : NDArray[Any]
        Global longitude coordinates.
    cropping_distance : float, optional
        Cropping distance. Defaults to 2.0.

    Returns
    -------
    NDArray[Any]
        Array of indices of the closest points.
    """
    from scipy.spatial import cKDTree

    assert global_lats.ndim == 1
    assert global_lons.ndim == 1
    assert lats.ndim == 1
    assert lons.ndim == 1

    assert global_lats.shape == global_lons.shape
    assert lats.shape == lons.shape

    north = np.amax(lats)
    south = np.amin(lats)
    east = np.amax(lons)
    west = np.amin(lons)

    # Reduce the global grid to the area of interest

    mask = cropping_mask(
        global_lats,
        global_lons,
        np.min([90.0, north + cropping_distance]),
        west - cropping_distance,
        np.max([-90.0, south - cropping_distance]),
        east + cropping_distance,
    )

    # return mask
    global_lats_masked = global_lats[mask]
    global_lons_masked = global_lons[mask]

    global_xyx = latlon_to_xyz(global_lats_masked, global_lons_masked)
    global_points = np.array(global_xyx).transpose()

    xyx = latlon_to_xyz(lats, lons)
    points = np.array(xyx).transpose()

    # Use a cKDTree to find the nearest points
    _, indices = cKDTree(points).query(global_points, k=1)

    return np.array([i for i in indices])


def outline(lats: NDArray[Any], lons: NDArray[Any], neighbours: int = 5) -> List[int]:
    """Find the outline of the grid points.

    Parameters
    ----------
    lats : NDArray[Any]
        Latitude coordinates.
    lons : NDArray[Any]
        Longitude coordinates.
    neighbours : int, optional
        Number of neighbours. Defaults to 5.

    Returns
    -------
    List[int]
        Indices of the outline points.
    """
    from scipy.spatial import cKDTree

    xyx = latlon_to_xyz(lats, lons)
    grid_points = np.array(xyx).transpose()

    # Use a cKDTree to find the nearest points
    _, indices = cKDTree(grid_points).query(grid_points, k=neighbours)

    # Centre of the Earth
    zero = np.array([0.0, 0.0, 0.0])

    outside = []

    for i, (point, index) in enumerate(zip(grid_points, indices)):
        inside = False
        for j in range(1, neighbours):
            t = Triangle3D(
                grid_points[index[j]],
                grid_points[index[(j + 1) % neighbours]],
                grid_points[index[(j + 2) % neighbours]],
            )
            inside = t.intersect(zero, point)
            if inside:
                break

        if not inside:
            outside.append(i)

    return outside


def deserialise_mask(encoded: str) -> NDArray[Any]:
    """Deserialise a mask from a base64 encoded string.

    Parameters
    ----------
    encoded : str
        Base64 encoded string.

    Returns
    -------
    NDArray[Any]
        Deserialised mask array.
    """
    import pickle
    import zlib

    packed = pickle.loads(zlib.decompress(base64.b64decode(encoded)))

    mask = []
    value = False
    for count in packed:
        mask.extend([value] * count)
        value = not value
    return np.array(mask, dtype=bool)


def _serialise_mask(mask: NDArray[Any]) -> str:
    """Serialise a mask to a base64 encoded string.

    Parameters
    ----------
    mask : NDArray[Any]
        Mask array.

    Returns
    -------
    str
        Base64 encoded string.
    """
    import pickle
    import zlib

    assert len(mask.shape) == 1
    assert len(mask)

    packed = []
    last = mask[0]
    count = 1

    for value in mask[1:]:
        if value == last:
            count += 1
        else:
            packed.append(count)
            last = value
            count = 1

    packed.append(count)

    # We always start with an 'off' value
    # So if the first value is 'on', we need to add a zero
    if mask[0]:
        packed.insert(0, 0)

    return base64.b64encode(zlib.compress(pickle.dumps(packed))).decode("utf-8")


def serialise_mask(mask: NDArray[Any]) -> str:
    """Serialise a mask and ensure it can be deserialised.

    Parameters
    ----------
    mask : NDArray[Any]
        Mask array.

    Returns
    -------
    str
        Base64 encoded string.
    """
    result = _serialise_mask(mask)
    # Make sure we can deserialise it
    assert np.all(mask == deserialise_mask(result))
    return result


def nearest_grid_points(
    source_latitudes: NDArray[Any],
    source_longitudes: NDArray[Any],
    target_latitudes: NDArray[Any],
    target_longitudes: NDArray[Any],
    max_distance: float = None,
) -> NDArray[Any]:
    """Find the nearest grid points from source to target coordinates.

    Parameters
    ----------
    source_latitudes : NDArray[Any]
        Source latitude coordinates.
    source_longitudes : NDArray[Any]
        Source longitude coordinates.
    target_latitudes : NDArray[Any]
        Target latitude coordinates.
    target_longitudes : NDArray[Any]
        Target longitude coordinates.
    max_distance: float, optional
        Maximum distance between nearest point and point to interpolate. Defaults to None.
        For example, 1e-3 is 1 km.

    Returns
    -------
    NDArray[Any]
        Indices of the nearest grid points.
    """
    # TODO: Use the one from anemoi.utils.grids instead
    # from anemoi.utils.grids import ...
    from scipy.spatial import cKDTree

    source_xyz = latlon_to_xyz(source_latitudes, source_longitudes)
    source_points = np.array(source_xyz).transpose()

    target_xyz = latlon_to_xyz(target_latitudes, target_longitudes)
    target_points = np.array(target_xyz).transpose()
    if max_distance is None:
        _, indices = cKDTree(source_points).query(target_points, k=1)
    else:
        _, indices = cKDTree(source_points).query(target_points, k=1, distance_upper_bound=max_distance)
    return indices


if __name__ == "__main__":
    global_lats, global_lons = np.meshgrid(
        np.linspace(90, -90, 90),
        np.linspace(-180, 180, 180),
    )
    global_lats = global_lats.flatten()
    global_lons = global_lons.flatten()

    lats, lons = np.meshgrid(
        np.linspace(50, 40, 100),
        np.linspace(-10, 15, 100),
    )
    lats = lats.flatten()
    lons = lons.flatten()

    mask = cutout_mask(lats, lons, global_lats, global_lons, cropping_distance=5.0)

    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(10, 5))
    plt.scatter(global_lons, global_lats, s=0.01, marker="o", c="r")
    plt.scatter(global_lons[mask], global_lats[mask], s=0.1, c="k")
    # plt.scatter(lons, lats, s=0.01)
    plt.savefig("cutout.png")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import os
import warnings
from contextlib import contextmanager
from typing import Any
from typing import Union

import numpy as np
from earthkit.data import settings
from numpy.typing import NDArray


def cache_context(dirname: str) -> contextmanager:
    """Context manager for setting a temporary cache directory.

    Parameters
    ----------
    dirname : str
        The directory name for the cache.

    Returns
    -------
    contextmanager
        A context manager that sets the cache directory.
    """

    @contextmanager
    def no_cache_context():
        yield

    if dirname is None:
        return no_cache_context()

    os.makedirs(dirname, exist_ok=True)
    # return settings.temporary("cache-directory", dirname)
    return settings.temporary({"cache-policy": "user", "user-cache-directory": dirname})


def to_datetime_list(*args: Any, **kwargs: Any) -> list[datetime.datetime]:
    """Convert various date formats to a list of datetime objects.

    Parameters
    ----------
    *args : Any
        Positional arguments for date conversion.
    **kwargs : Any
        Keyword arguments for date conversion.

    Returns
    -------
    list[datetime.datetime]
        A list of datetime objects.
    """
    from earthkit.data.utils.dates import to_datetime_list as to_datetime_list_

    warnings.warn(
        "to_datetime_list() is deprecated. Call earthkit.data.utils.dates.to_datetime_list() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return to_datetime_list_(*args, **kwargs)


def to_datetime(*args: Any, **kwargs: Any) -> datetime.datetime:
    """Convert various date formats to a single datetime object.

    Parameters
    ----------
    *args : Any
        Positional arguments for date conversion.
    **kwargs : Any
        Keyword arguments for date conversion.

    Returns
    -------
    datetime.datetime
        A datetime object.
    """
    from earthkit.data.utils.dates import to_datetime as to_datetime_

    warnings.warn(
        "to_datetime() is deprecated. Call earthkit.data.utils.dates.to_datetime() instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    return to_datetime_(*args, **kwargs)


def make_list_int(value: Union[str, list, tuple, int]) -> list[int]:
    """Convert a string, list, tuple, or integer to a list of integers.

    Parameters
    ----------
    value : str or list or tuple or int
        The value to convert.

    Returns
    -------
    list[int]
        A list of integers.

    Raises
    ------
    ValueError
        If the value cannot be converted to a list of integers.
    """
    # Convert a string like "1/2/3" or "1/to/3" or "1/to/10/by/2" to a list of integers.
    # Moved to anemoi.utils.humanize
    # replace with from anemoi.utils.humanize import make_list_int
    # when anemoi-utils is released and pyproject.toml is updated
    if isinstance(value, str):
        if "/" not in value:
            return [value]
        bits = value.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            value = list(range(int(bits[0]), int(bits[2]) + 1, 1))

        elif len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            value = list(range(int(bits[0]), int(bits[2]) + int(bits[4]), int(bits[4])))

    if isinstance(value, list):
        return value
    if isinstance(value, tuple):
        return value
    if isinstance(value, int):
        return [value]

    raise ValueError(f"Cannot make list from {value}")


def normalize_and_check_dates(
    dates: list[datetime.datetime],
    start: datetime.datetime,
    end: datetime.datetime,
    frequency: datetime.timedelta,
    dtype: str = "datetime64[s]",
) -> NDArray[Any]:
    """Normalize and check a list of dates against a specified frequency.

    Parameters
    ----------
    dates : list[datetime.datetime]
        The list of dates to check.
    start : datetime.datetime
        The start date.
    end : datetime.datetime
        The end date.
    frequency : datetime.timedelta
        The frequency of the dates.
    dtype : str, optional
        The data type of the dates, by default "datetime64[s]".

    Returns
    -------
    NDArray[Any]
        An array of normalized dates.

    Raises
    ------
    ValueError
        If the final date size does not match the data shape.
    """
    dates = [d.hdate if hasattr(d, "hdate") else d for d in dates]

    assert isinstance(frequency, datetime.timedelta), frequency
    start = np.datetime64(start)
    end = np.datetime64(end)
    delta = np.timedelta64(frequency)

    res = []
    while start <= end:
        res.append(start)
        start += delta
    dates_ = np.array(res).astype(dtype)

    if len(dates_) != len(dates):
        raise ValueError(
            f"Final date size {len(dates_)} (from {dates_[0]} to {dates_[-1]}, "
            f"{frequency=}) does not match data shape {len(dates)} (from {dates[0]} to "
            f"{dates[-1]})."
        )

    for i, (d1, d2) in enumerate(zip(dates, dates_)):
        d1 = np.datetime64(d1)
        d2 = np.datetime64(d2)
        assert d1 == d2, (i, d1, d2)

    return dates_



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Optional

import earthkit.data as ekd
import tqdm
from anemoi.transform.fields import new_field_from_numpy
from anemoi.transform.fields import new_fieldlist_from_list
from anemoi.utils.humanize import plural
from earthkit.geo.rotate import rotate_vector

from .legacy import legacy_filter


@legacy_filter(__file__)
def execute(
    context: Any,
    input: ekd.FieldList,
    x_wind: str,
    y_wind: str,
    source_projection: Optional[str] = None,
    target_projection: str = "+proj=longlat",
) -> ekd.FieldList:
    """Rotate wind components from one projection to another.

    Parameters
    ----------
    context : Any
        The context in which the function is executed.
    input : ekd.FieldList
        List of input fields.
    x_wind : str
        X wind component parameter.
    y_wind : str
        Y wind component parameter.
    source_projection : Optional[str], optional
        Source projection, by default None.
    target_projection : str, optional
        Target projection, by default "+proj=longlat".

    Returns
    -------
    ekd.FieldList
        Array of fields with rotated wind components.
    """
    from pyproj import CRS

    context.trace("🔄", "Rotating winds (extracting winds from ", plural(len(input), "field"))

    result = []

    wind_params: tuple[str, str] = (x_wind, y_wind)
    wind_pairs: Dict[tuple, Dict[str, Any]] = defaultdict(dict)

    for f in input:
        key = f.metadata(namespace="mars")
        param = key.pop("param")

        if param not in wind_params:
            result.append(f)
            continue

        key = tuple(key.items())

        if param in wind_pairs[key]:
            raise ValueError(f"Duplicate wind component {param} for {key}")

        wind_pairs[key][param] = f

    context.trace("🔄", "Rotating", plural(len(wind_pairs), "wind"), "(speed will likely include data download)")

    for _, pairs in tqdm.tqdm(list(wind_pairs.items())):
        if len(pairs) != 2:
            raise ValueError("Missing wind component")

        x = pairs[x_wind]
        y = pairs[y_wind]

        assert x.grid_mapping == y.grid_mapping

        lats, lons = x.grid_points()
        x_new, y_new = rotate_vector(
            lats,
            lons,
            x.to_numpy(flatten=True),
            y.to_numpy(flatten=True),
            (source_projection if source_projection is not None else CRS.from_cf(x.grid_mapping)),
            target_projection,
        )

        result.append(new_field_from_numpy(x, x_new))
        result.append(new_field_from_numpy(y, y_new))

    return new_fieldlist_from_list(result)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from abc import abstractmethod
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from numpy.typing import NDArray

from ..grids import nearest_grid_points
from .dataset import Dataset
from .dataset import FullIndex
from .dataset import Shape
from .dataset import TupleIndex
from .debug import Node
from .forwards import Combined
from .indexing import apply_index_to_slices_changes
from .indexing import index_to_slices
from .indexing import update_tuple
from .misc import _auto_adjust
from .misc import _open

LOG = logging.getLogger(__name__)


class Complement(Combined):
    """A class to complement a target dataset with variables from a source dataset,
    interpolated on the grid of the target dataset.

    Attributes
    ----------
    target : Dataset
        The target dataset.
    source : Dataset
        The source dataset.
    variables : List[str]
        List of variables to be added to the target dataset.
    """

    def __init__(
        self,
        target: Dataset,
        source: Dataset,
        what: str = "variables",
        interpolation: str = "nearest",
    ) -> None:
        """Initializes the Complement class.

        Parameters
        ----------
        target : Dataset
            The target dataset.
        source : Dataset
            The source dataset.
        what : str, optional
            What to complement, default is "variables".
        interpolation : str, optional
            Interpolation method, default is "nearest".
        """
        super().__init__([target, source])

        # We had the variables of dataset[1] to dataset[0]
        # interpoated on the grid of dataset[0]

        self._target: Dataset = target
        self._source: Dataset = source

        self._variables = []

        # Keep the same order as the original dataset
        for v in self._source.variables:
            if v not in self._target.variables:
                self._variables.append(v)

        if not self._variables:
            raise ValueError("Augment: no missing variables")

    @property
    def variables(self) -> List[str]:
        """Returns the list of variables to be added to the target dataset."""
        return self._variables

    @property
    def name_to_index(self) -> Dict[str, int]:
        """Returns a dictionary mapping variable names to their indices."""
        return {v: i for i, v in enumerate(self.variables)}

    @property
    def shape(self) -> Shape:
        """Returns the shape of the complemented dataset."""
        shape = self._target.shape
        return (shape[0], len(self._variables)) + shape[2:]

    @property
    def variables_metadata(self) -> Dict[str, Any]:
        """Returns the metadata of the variables to be added to the target dataset."""
        return {k: v for k, v in self._source.variables_metadata.items() if k in self._variables}

    def check_same_variables(self, d1: Dataset, d2: Dataset) -> None:
        """Checks if the variables in two datasets are the same.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        pass

    @cached_property
    def missing(self) -> Set[int]:
        """Returns the set of missing indices in the source and target datasets."""
        missing = self._source.missing.copy()
        missing = missing | self._target.missing
        return set(missing)

    def tree(self) -> Node:
        """Generates a hierarchical tree structure for the Complement instance and its associated datasets.

        Returns
        -------
        Node
            A Node object representing the Complement instance as the root node, with each dataset in self.datasets represented as a child node.
        """
        return Node(self, [d.tree() for d in (self._target, self._source)])

    def __getitem__(self, index: FullIndex) -> NDArray[Any]:
        """Gets the data at the specified index.

        Parameters
        ----------
        index : FullIndex
            The index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data at the specified index.
        """
        if isinstance(index, (int, slice)):
            index = (index, slice(None), slice(None), slice(None))
        return self._get_tuple(index)

    @abstractmethod
    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Abstract method to get the data at the specified tuple index.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data at the specified tuple index.
        """
        pass


class ComplementNone(Complement):
    """A class to complement a target dataset with variables from a source dataset without interpolation."""

    def __init__(self, target: Any, source: Any) -> None:
        """Initializes the ComplementNone class.

        Parameters
        ----------
        target : Any
            The target dataset.
        source : Any
            The source dataset.
        """
        super().__init__(target, source)

    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Gets the data at the specified tuple index without interpolation.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data at the specified tuple index.
        """
        index, changes = index_to_slices(index, self.shape)
        result = self._source[index]
        return apply_index_to_slices_changes(result, changes)


class ComplementNearest(Complement):
    """A class to complement a target dataset with variables from a source dataset using nearest neighbor interpolation."""

    def __init__(self, target: Any, source: Any, max_distance: float = None) -> None:
        """Initializes the ComplementNearest class.

        Parameters
        ----------
        target : Any
            The target dataset.
        source : Any
            The source dataset.
        max_distance : float, optional
            The maximum distance for nearest neighbor interpolation, default is None.
        """
        super().__init__(target, source)

        self._nearest_grid_points = nearest_grid_points(
            self._source.latitudes,
            self._source.longitudes,
            self._target.latitudes,
            self._target.longitudes,
            max_distance=max_distance,
        )

    def check_compatibility(self, d1: Dataset, d2: Dataset) -> None:
        """Checks the compatibility of two datasets for nearest neighbor interpolation.

        Parameters
        ----------
        d1 : Dataset
            The first dataset.
        d2 : Dataset
            The second dataset.
        """
        pass

    def _get_tuple(self, index: TupleIndex) -> NDArray[Any]:
        """Gets the data at the specified tuple index using nearest neighbor interpolation.

        Parameters
        ----------
        index : TupleIndex
            The tuple index to retrieve data from.

        Returns
        -------
        NDArray[Any]
            The data at the specified tuple index.
        """
        variable_index = 1
        index, changes = index_to_slices(index, self.shape)
        index, previous = update_tuple(index, variable_index, slice(None))
        source_index = [self._source.name_to_index[x] for x in self.variables[previous]]
        source_data = self._source[index[0], source_index, index[2], ...]
        target_data = source_data[..., self._nearest_grid_points]

        result = target_data[..., index[3]]

        return apply_index_to_slices_changes(result, changes)


def complement_factory(args: Tuple, kwargs: dict) -> Dataset:
    """Factory function to create a Complement instance based on the provided arguments.

    Parameters
    ----------
    args : Tuple
        Positional arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    Dataset
        The complemented dataset.
    """
    from .select import Select

    assert len(args) == 0, args

    source = kwargs.pop("source")
    target = kwargs.pop("complement")
    what = kwargs.pop("what", "variables")
    interpolation = kwargs.pop("interpolation", "none")

    if what != "variables":
        raise NotImplementedError(f"Complement what={what} not implemented")

    if interpolation not in ("none", "nearest"):
        raise NotImplementedError(f"Complement method={interpolation} not implemented")

    source = _open(source)
    target = _open(target)
    # `select` is the same as `variables`
    (source, target), kwargs = _auto_adjust([source, target], kwargs, exclude=["select"])

    Class = {
        None: ComplementNone,
        "none": ComplementNone,
        "nearest": ComplementNearest,
    }[interpolation]

    complement = Class(target=target, source=source)._subset(**kwargs)

    # Will join the datasets along the variables axis
    reorder = source.variables
    complemented = _open([target, complement])
    ordered = (
        Select(complemented, complemented._reorder_to_columns(reorder), {"reoder": reorder})._subset(**kwargs).mutate(),
    )
    return ordered



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import os
from copy import deepcopy
from typing import Any
from typing import Optional
from typing import Union

import yaml
from anemoi.utils.config import DotDict
from anemoi.utils.config import load_any_dict_format
from earthkit.data.core.order import normalize_order_by

from anemoi.datasets.dates.groups import Groups

LOG = logging.getLogger(__name__)


def _get_first_key_if_dict(x: Union[str, dict]) -> str:
    """Returns the first key if the input is a dictionary, otherwise returns the input string.

    Parameters
    ----------
    x : str or dict
        Input string or dictionary.

    Returns
    -------
    str
        The first key if input is a dictionary, otherwise the input string.
    """
    if isinstance(x, str):
        return x
    return list(x.keys())[0]


def ensure_element_in_list(lst: list, elt: str, index: int) -> list:
    """Ensures that a specified element is present at a given index in a list.

    Parameters
    ----------
    lst : list
        The list to check.
    elt : str
        The element to ensure is in the list.
    index : int
        The index at which the element should be present.

    Returns
    -------
    list
        The modified list with the element at the specified index.
    """
    if elt in lst:
        assert lst[index] == elt
        return lst

    _lst = [_get_first_key_if_dict(d) for d in lst]
    if elt in _lst:
        assert _lst[index] == elt
        return lst

    return lst[:index] + [elt] + lst[index:]


def check_dict_value_and_set(dic: dict, key: str, value: Any) -> None:
    """Checks if a dictionary contains a specific key-value pair and sets it if not present.

    Parameters
    ----------
    dic : dict
        The dictionary to check.
    key : str
        The key to check in the dictionary.
    value : Any
        The value to set if the key is not present.

    Raises
    ------
    ValueError
        If the key is present but with a different value.
    """
    if key in dic:
        if dic[key] == value:
            return
        raise ValueError(f"Cannot use {key}={dic[key]}. Must use {value}.")
    LOG.info(f"Setting {key}={value} in config")
    dic[key] = value


def resolve_includes(config: Union[dict, list]) -> Union[dict, list]:
    """Resolves '<<' includes in a configuration dictionary or list.

    Parameters
    ----------
    config : dict or list
        The configuration to resolve includes for.

    Returns
    -------
    dict or list
        The configuration with includes resolved.
    """
    if isinstance(config, list):
        return [resolve_includes(c) for c in config]
    if isinstance(config, dict):
        include = config.pop("<<", {})
        new = deepcopy(include)
        new.update(config)
        return {k: resolve_includes(v) for k, v in new.items()}
    return config


class Config(DotDict):
    """Configuration class that extends DotDict to handle configuration loading and processing."""

    def __init__(self, config: Optional[Union[str, dict]] = None, **kwargs):
        """Initializes the Config object.

        Parameters
        ----------
        config : str or dict, optional
            Path to the configuration file or a dictionary. Defaults to None.
        **kwargs
            Additional keyword arguments to update the configuration.
        """
        if isinstance(config, str):
            self.config_path = os.path.realpath(config)
            config = load_any_dict_format(config)
        else:
            config = deepcopy(config if config is not None else {})
        config = resolve_includes(config)
        config.update(kwargs)
        super().__init__(config)


class OutputSpecs:
    """Class to handle output specifications for datasets."""

    def __init__(self, config: Config, parent: Any):
        """Initializes the OutputSpecs object.

        Parameters
        ----------
        config : Config
            The configuration object.
        parent : Any
            The parent object.
        """
        self.config = config
        if "order_by" in config:
            assert isinstance(config.order_by, dict), config.order_by

        self.parent = parent

    @property
    def dtype(self) -> str:
        """Returns the data type for the output."""
        return self.config.dtype

    @property
    def order_by_as_list(self) -> list[dict]:
        """Returns the order_by configuration as a list of dictionaries."""
        return [{k: v} for k, v in self.config.order_by.items()]

    def get_chunking(self, coords: dict) -> tuple:
        """Returns the chunking configuration based on coordinates.

        Parameters
        ----------
        coords : dict
            The coordinates dictionary.

        Returns
        -------
        tuple
            The chunking configuration.
        """
        user = deepcopy(self.config.chunking)
        chunks = []
        for k, v in coords.items():
            if k in user:
                chunks.append(user.pop(k))
            else:
                chunks.append(len(v))
        if user:
            raise ValueError(
                f"Unused chunking keys from config: {list(user.keys())}, not in known keys : {list(coords.keys())}"
            )
        return tuple(chunks)

    @property
    def order_by(self) -> dict:
        """Returns the order_by configuration."""
        return self.config.order_by

    @property
    def remapping(self) -> dict:
        """Returns the remapping configuration."""
        return self.config.remapping

    @property
    def flatten_grid(self) -> bool:
        """Returns whether the grid should be flattened."""
        return self.config.flatten_grid

    @property
    def statistics(self) -> str:
        """Returns the statistics configuration."""
        return self.config.statistics


class LoadersConfig(Config):
    """Configuration class for dataset loaders."""

    def __init__(self, config: dict, *args, **kwargs):
        """Initializes the LoadersConfig object.

        Parameters
        ----------
        config : dict
            The configuration dictionary.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.
        """
        super().__init__(config, *args, **kwargs)

        # TODO: should use a json schema to validate the config

        self.setdefault("dataset_status", "experimental")
        self.setdefault("description", "No description provided.")
        self.setdefault("licence", "unknown")
        self.setdefault("attribution", "unknown")

        self.setdefault("build", Config())
        self.build.setdefault("group_by", "monthly")
        self.build.setdefault("use_grib_paramid", False)
        self.build.setdefault("variable_naming", "default")
        variable_naming = dict(
            param="{param}",
            param_levelist="{param}_{levelist}",
            default="{param}_{levelist}",
        ).get(self.build.variable_naming, self.build.variable_naming)

        self.setdefault("output", Config())
        self.output.setdefault("order_by", ["valid_datetime", "param_level", "number"])
        self.output.setdefault("remapping", Config(param_level=variable_naming))
        self.output.setdefault("statistics", "param_level")
        self.output.setdefault("chunking", Config(dates=1, ensembles=1))
        self.output.setdefault("dtype", "float32")

        if "statistics_start" in self.output:
            raise ValueError("statistics_start is not supported anymore. Use 'statistics:start:' instead")
        if "statistics_end" in self.output:
            raise ValueError("statistics_end is not supported anymore. Use 'statistics:end:' instead")

        self.setdefault("statistics", Config())
        if "allow_nans" not in self.statistics:
            self.statistics.allow_nans = []

        check_dict_value_and_set(self.output, "flatten_grid", True)
        check_dict_value_and_set(self.output, "ensemble_dimension", 2)

        assert isinstance(self.output.order_by, (list, tuple)), self.output.order_by
        self.output.order_by = ensure_element_in_list(self.output.order_by, "number", self.output.ensemble_dimension)

        order_by = self.output.order_by
        assert len(order_by) == 3, order_by
        assert _get_first_key_if_dict(order_by[0]) == "valid_datetime", order_by
        assert _get_first_key_if_dict(order_by[2]) == "number", order_by

        self.output.order_by = normalize_order_by(self.output.order_by)

        self.dates["group_by"] = self.build.group_by

        ###########

        self.reading_chunks = self.get("reading_chunks")

    def get_serialisable_dict(self) -> dict:
        """Returns a serializable dictionary representation of the configuration.

        Returns
        -------
        dict
            The serializable dictionary.
        """
        return _prepare_serialisation(self)


def _prepare_serialisation(o: Any) -> Any:
    """Prepares an object for serialization.

    Parameters
    ----------
    o : Any
        The object to prepare.

    Returns
    -------
    Any
        The prepared object.
    """
    if isinstance(o, dict):
        dic = {}
        for k, v in o.items():
            v = _prepare_serialisation(v)
            if k == "order_by" and isinstance(v, dict):
                # zarr attributes are saved with sort_keys=True
                # and ordered dict are reordered.
                # This is a problem for "order_by"
                # We ensure here that the order_by key contains
                # a list of dict
                v = [{kk: vv} for kk, vv in v.items()]
            dic[k] = v
        return dic

    if isinstance(o, (list, tuple)):
        return [_prepare_serialisation(v) for v in o]

    if o in (None, True, False):
        return o

    if isinstance(o, (str, int, float)):
        return o

    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()

    return str(o)


def set_to_test_mode(cfg: dict) -> None:
    """Modifies the configuration to run in test mode.

    Parameters
    ----------
    cfg : dict
        The configuration dictionary.
    """
    NUMBER_OF_DATES = 4

    LOG.warning(f"Running in test mode. Changing the list of dates to use only {NUMBER_OF_DATES}.")
    groups = Groups(**LoadersConfig(cfg).dates)

    dates = groups.provider.values
    cfg["dates"] = dict(
        start=dates[0],
        end=dates[NUMBER_OF_DATES - 1],
        frequency=groups.provider.frequency,
        group_by=NUMBER_OF_DATES,
    )

    def set_element_to_test(obj):
        if isinstance(obj, (list, tuple)):
            for v in obj:
                set_element_to_test(v)
            return
        if isinstance(obj, (dict, DotDict)):
            if "grid" in obj:
                previous = obj["grid"]
                obj["grid"] = "20./20."
                LOG.warning(f"Running in test mode. Setting grid to {obj['grid']} instead of {previous}")
            if "number" in obj:
                if isinstance(obj["number"], (list, tuple)):
                    previous = obj["number"]
                    obj["number"] = previous[0:3]
                    LOG.warning(f"Running in test mode. Setting number to {obj['number']} instead of {previous}")
            for k, v in obj.items():
                set_element_to_test(v)
            if "constants" in obj:
                constants = obj["constants"]
                if "param" in constants and isinstance(constants["param"], list):
                    constants["param"] = ["cos_latitude"]

    set_element_to_test(cfg)


def loader_config(config: dict, is_test: bool = False) -> LoadersConfig:
    """Loads and validates the configuration for dataset loaders.

    Parameters
    ----------
    config : dict
        The configuration dictionary.
    is_test : bool, optional
        Whether to run in test mode. Defaults to False.

    Returns
    -------
    LoadersConfig
        The validated configuration object.
    """
    config = Config(config)
    if is_test:
        set_to_test_mode(config)
    obj = LoadersConfig(config)

    # yaml round trip to check that serialisation works as expected
    copy = obj.get_serialisable_dict()
    copy = yaml.load(yaml.dump(copy), Loader=yaml.SafeLoader)
    copy = Config(copy)
    copy = LoadersConfig(config)

    a = yaml.dump(obj)
    b = yaml.dump(copy)
    if a != b:
        print(a)
        print(b)
        raise ValueError("Serialisation failed")

    if "env" in copy:
        for k, v in copy["env"].items():
            LOG.info(f"Setting env variable {k}={v}")
            os.environ[k] = str(v)

    return copy


def build_output(*args, **kwargs) -> OutputSpecs:
    """Builds the output specifications.

    Parameters
    ----------
    *args
        Additional positional arguments.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    OutputSpecs
        The output specifications object.
    """
    return OutputSpecs(*args, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from copy import deepcopy
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from earthkit.data import FieldList

from anemoi.datasets.dates import DatesProvider

from ...dates.groups import GroupOfDates
from .action import Action
from .action import action_factory
from .empty import EmptyResult
from .misc import _tidy
from .misc import assert_fieldlist
from .result import Result
from .template import notify_result
from .trace import trace_datasource
from .trace import trace_select

LOG = logging.getLogger(__name__)


class ConcatResult(Result):
    """Represents the result of concatenating multiple results."""

    def __init__(
        self,
        context: object,
        action_path: List[str],
        group_of_dates: GroupOfDates,
        results: List[Result],
        **kwargs: Any,
    ) -> None:
        """Initializes a ConcatResult instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : List[str]
            The action path.
        group_of_dates : GroupOfDates
            The group of dates.
        results : List[Result]
            The list of results.
        kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, action_path, group_of_dates)
        self.results = [r for r in results if not r.empty]

    @cached_property
    @assert_fieldlist
    @notify_result
    @trace_datasource
    def datasource(self) -> FieldList:
        """Returns the concatenated datasource from all results."""
        ds = EmptyResult(self.context, self.action_path, self.group_of_dates).datasource
        for i in self.results:
            ds += i.datasource
        return _tidy(ds)

    @property
    def variables(self) -> List[str]:
        """Returns the list of variables, ensuring all results have the same variables."""
        variables = None
        for f in self.results:
            if f.empty:
                continue
            if variables is None:
                variables = f.variables
            assert variables == f.variables, (variables, f.variables)
        assert variables is not None, self.results
        return variables

    def __repr__(self) -> str:
        """Returns a string representation of the ConcatResult instance.

        Returns
        -------
        str
            A string representation of the ConcatResult instance.
        """
        content = "\n".join([str(i) for i in self.results])
        return self._repr(content)


class ConcatAction(Action):
    """Represents an action that concatenates multiple actions based on their dates."""

    def __init__(self, context: object, action_path: List[str], *configs: Dict[str, Any]) -> None:
        """Initializes a ConcatAction instance.

        Parameters
        ----------
        context : object
            The context object.
        action_path : List[str]
            The action path.
        configs : Dict[str, Any]
            The configuration dictionaries.
        """
        super().__init__(context, action_path, *configs)
        parts = []
        for i, cfg in enumerate(configs):
            if "dates" not in cfg:
                raise ValueError(f"Missing 'dates' in {cfg}")
            cfg = deepcopy(cfg)
            dates_cfg = cfg.pop("dates")
            assert isinstance(dates_cfg, dict), dates_cfg
            filtering_dates = DatesProvider.from_config(**dates_cfg)
            action = action_factory(cfg, context, action_path + [str(i)])
            parts.append((filtering_dates, action))
        self.parts = parts

    def __repr__(self) -> str:
        """Returns a string representation of the ConcatAction instance.

        Returns
        -------
        str
            A string representation of the ConcatAction instance.
        """
        content = "\n".join([str(i) for i in self.parts])
        return self._repr(content)

    @trace_select
    def select(self, group_of_dates: GroupOfDates) -> Union[ConcatResult, EmptyResult]:
        """Selects the concatenated result for the given group of dates.

        Parameters
        ----------
        group_of_dates : GroupOfDates
            The group of dates.

        Returns
        -------
        Union[ConcatResult, EmptyResult]
            The concatenated result or an empty result.
        """
        from anemoi.datasets.dates.groups import GroupOfDates

        results = []
        for filtering_dates, action in self.parts:
            newdates = GroupOfDates(sorted(set(group_of_dates) & set(filtering_dates)), group_of_dates.provider)
            if newdates:
                results.append(action.select(newdates))
        if not results:
            return EmptyResult(self.context, self.action_path, group_of_dates)

        return ConcatResult(self.context, self.action_path, group_of_dates, results)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
import math
from functools import cached_property
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import xarray as xr

from .field import XArrayField

LOG = logging.getLogger(__name__)


class Variable:
    """Represents a variable in an xarray dataset.

    Attributes
    ----------
    ds : xr.Dataset
        The xarray dataset.
    variable : xr.DataArray
        The data array representing the variable.
    coordinates : List[Any]
        List of coordinates associated with the variable.
    grid : Any
        The grid associated with the variable.
    time : Any
        The time dimension associated with the variable.
    metadata : Dict[str, Any]
        Metadata associated with the variable.
    """

    def __init__(
        self,
        *,
        ds: xr.Dataset,
        variable: xr.DataArray,
        coordinates: List[Any],
        grid: Any,
        time: Any,
        metadata: Dict[str, Any],
    ):
        """Initialize the Variable object.

        Parameters
        ----------
        ds : xr.Dataset
            The xarray dataset.
        variable : xr.DataArray
            The data array representing the variable.
        coordinates : List[Any]
            List of coordinates associated with the variable.
        grid : Any
            The grid associated with the variable.
        time : Any
            The time dimension associated with the variable.
        metadata : Dict[str, Any]
            Metadata associated with the variable.
        """
        self.ds = ds
        self.variable = variable

        self.grid = grid
        self.coordinates = coordinates

        self._metadata = metadata.copy()
        self._metadata.update({"variable": variable.name, "param": variable.name})

        self.time = time

        self.shape = tuple(len(c.variable) for c in coordinates if c.is_dim and not c.scalar and not c.is_grid)
        self.names = {c.variable.name: c for c in coordinates if c.is_dim and not c.scalar and not c.is_grid}
        self.by_name = {c.variable.name: c for c in coordinates}

        # We need that alias for the time dimension
        self._aliases = dict(valid_datetime="time")

        self.length = math.prod(self.shape)

    @property
    def name(self) -> str:
        """Return the name of the variable."""
        return str(self.variable.name)

    def __len__(self) -> int:
        """Return the length of the variable.

        Returns
        -------
        int
            The length of the variable.
        """
        return self.length

    @property
    def grid_mapping(self) -> Optional[Dict[str, Any]]:
        """Return the grid mapping of the variable."""
        grid_mapping = self.variable.attrs.get("grid_mapping", None)
        if grid_mapping is None:
            return None
        return self.ds[grid_mapping].attrs

    def grid_points(self) -> Any:
        """Return the grid points of the variable.

        Returns
        -------
        Any
            The grid points of the variable.
        """
        return self.grid.grid_points

    @property
    def latitudes(self) -> Any:
        """Return the latitudes of the variable."""
        return self.grid.latitudes

    @property
    def longitudes(self) -> Any:
        """Return the longitudes of the variable."""
        return self.grid.longitudes

    def __repr__(self) -> str:
        """Return a string representation of the variable.

        Returns
        -------
        str
            A string representation of the variable.
        """
        return "Variable[name=%s,coordinates=%s,metadata=%s]" % (
            self.variable.name,
            self.coordinates,
            self._metadata,
        )

    def __getitem__(self, i: int) -> "XArrayField":
        """Get a 2D field from the variable.

        Parameters
        ----------
        i : int
            Index of the field.

        Returns
        -------
        XArrayField
            The 2D field at the specified index.

        Raises
        ------
        IndexError
            If the index is out of range.
        """
        if i >= self.length:
            raise IndexError(i)

        coords = np.unravel_index(i, self.shape)
        kwargs = {k: v for k, v in zip(self.names, coords)}
        return XArrayField(self, self.variable.isel(kwargs))

    def sel(self, missing: Dict[str, Any], **kwargs: Any) -> Optional["Variable"]:
        """Select a subset of the variable based on the given coordinates.

        Parameters
        ----------
        missing : Dict[str, Any]
            Dictionary to store missing coordinates.
        **kwargs : Any
            Coordinates to select.

        Returns
        -------
        Optional[Variable]
            The selected subset of the variable.
        """
        if not kwargs:
            return self

        k, v = kwargs.popitem()

        user_provided_k = k

        if k == "valid_datetime":
            # Ask the Time object to select the valid datetime
            k = self.time.select_valid_datetime(self)
            if k is None:
                return None

        c = self.by_name.get(k)

        # assert c is not None, f"Could not find coordinate {k} in {self.variable.name} {self.coordinates} {list(self.by_name)}"

        if c is None:
            missing[k] = v
            return self.sel(missing, **kwargs)

        i = c.index(v)
        if i is None:
            if k != user_provided_k:
                LOG.warning(f"Could not find {user_provided_k}={v} in {c} (alias of {k})")
            else:
                LOG.warning(f"Could not find {k}={v} in {c}")
            return None

        coordinates = [x.reduced(i) if c is x else x for x in self.coordinates]

        metadata = self._metadata.copy()
        metadata.update({k: v})

        variable = Variable(
            ds=self.ds,
            variable=self.variable.isel({k: i}),
            coordinates=coordinates,
            grid=self.grid,
            time=self.time,
            metadata=metadata,
        )

        return variable.sel(missing, **kwargs)

    def match(self, **kwargs: Any) -> Tuple[bool, Optional[Dict[str, Any]]]:
        """Match the variable based on the given metadata.

        Parameters
        ----------
        **kwargs : Any
            Metadata to match.

        Returns
        -------
        Tuple[bool, Optional[Dict[str, Any]]]
            A tuple containing a boolean indicating if the match was successful and the remaining metadata.
        """
        if "param" in kwargs:
            assert "variable" not in kwargs
            kwargs["variable"] = kwargs.pop("param")

        if "variable" in kwargs:
            name = kwargs.pop("variable")
            if not isinstance(name, (list, tuple)):
                name = [name]
            if self.variable.name not in name:
                return False, None
            return True, kwargs
        return True, kwargs


class FilteredVariable:
    """Represents a filtered variable based on metadata.

    Attributes
    ----------
    variable : Variable
        The variable to filter.
    kwargs : Any
        Metadata to filter the variable.
    """

    def __init__(self, variable: Variable, **kwargs: Any):
        """Initialize the FilteredVariable object.

        Parameters
        ----------
        variable : Variable
            The variable to filter.
        **kwargs : Any
            Metadata to filter the variable.
        """
        self.variable = variable
        self.kwargs = kwargs

    @cached_property
    def fields(self) -> List["XArrayField"]:
        """Filter the fields of a variable based on metadata."""
        return [
            field
            for field in self.variable
            if all(field.metadata(k, default=None) == v for k, v in self.kwargs.items())
        ]

    @property
    def length(self) -> int:
        """Return the length of the filtered variable."""
        return len(self.fields)

    def __len__(self) -> int:
        """Return the length of the filtered variable.

        Returns
        -------
        int
            The length of the filtered variable.
        """
        return self.length

    def __getitem__(self, i: int) -> "XArrayField":
        """Get a field from the filtered variable.

        Parameters
        ----------
        i : int
            Index of the field.

        Returns
        -------
        XArrayField
            The field at the specified index.

        Raises
        ------
        IndexError
            If the index is out of range.
        """
        if i >= self.length:
            raise IndexError(i)
        return self.fields[i]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import glob
from typing import Any
from typing import Generator
from typing import List
from typing import Tuple

from earthkit.data.utils.patterns import Pattern


def _expand(paths: List[str]) -> Generator[str, None, None]:
    """Expand the given paths to include all matching file paths.

    Parameters
    ----------
    paths : List[str]
        List of paths to expand.

    Returns
    -------
    Generator[str]
        Expanded file paths.
    """
    if not isinstance(paths, list):
        paths = [paths]

    for path in paths:
        if path.startswith("file://"):
            path = path[7:]

        if path.startswith("http://"):
            yield path
            continue

        if path.startswith("https://"):
            yield path
            continue

        cnt = 0
        for p in glob.glob(path):
            yield p
            cnt += 1
        if cnt == 0:
            yield path


def iterate_patterns(
    path: str, dates: List[datetime.datetime], **kwargs: Any
) -> Generator[Tuple[str, List[str]], None, None]:
    """Iterate over patterns and expand them with given dates and additional keyword arguments.

    Parameters
    ----------
    path : str
        The pattern path to iterate over.
    dates : List[datetime.datetime]
        List of datetime objects to substitute in the pattern.
    **kwargs : Any
        Additional keyword arguments to substitute in the pattern.

    Returns
    -------
    Generator[Tuple[str, List[str]]]
        The expanded path and list of ISO formatted dates.
    """
    given_paths = path if isinstance(path, list) else [path]

    dates = [d.isoformat() for d in dates]
    if len(dates) > 0:
        kwargs["date"] = dates

    for path in given_paths:
        paths = Pattern(path, ignore_missing_keys=True).substitute(**kwargs)
        for path in _expand(paths):
            yield path, dates
